"""Robust pool for managing WebTransportStream objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pywebtransport.pool._base import _AsyncObjectPool
from pywebtransport.stream.stream import WebTransportStream

if TYPE_CHECKING:
    from pywebtransport.session.session import WebTransportSession


__all__: list[str] = ["StreamPool"]


class StreamPool(_AsyncObjectPool[WebTransportStream]):
    """A robust pool for reusing and managing concurrent bidirectional streams."""

    def __init__(self, *, session: WebTransportSession, max_size: int) -> None:
        """Initialize the stream pool."""

        async def factory() -> WebTransportStream:
            return await session.create_bidirectional_stream()

        super().__init__(max_size=max_size, factory=factory)

    async def _dispose(self, stream: WebTransportStream) -> None:
        """Close and dispose of a single stream object."""
        if not stream.is_closed:
            await stream.close()



"""Handle connection-level logic for the protocol engine."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pywebtransport import constants
from pywebtransport._protocol.events import (
    CleanupH3Stream,
    CloseQuicConnection,
    CompleteUserFuture,
    ConnectionClose,
    CreateH3Session,
    Effect,
    EmitConnectionEvent,
    EmitSessionEvent,
    FailUserFuture,
    GoawayReceived,
    HeadersReceived,
    InternalCleanupResources,
    SendH3Capsule,
    SendH3Goaway,
    SendH3Headers,
    TransportConnectionTerminated,
    TransportQuicParametersReceived,
    UserConnectionGracefulClose,
    UserCreateSession,
    UserGetConnectionDiagnostics,
)
from pywebtransport._protocol.state import ProtocolState, SessionStateData
from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import ConnectionError, ProtocolError
from pywebtransport.types import ConnectionId, ConnectionState, EventType, SessionState, StreamState
from pywebtransport.utils import generate_session_id, get_logger, get_timestamp

if TYPE_CHECKING:
    from pywebtransport.config import ClientConfig, ServerConfig


__all__: list[str] = []

logger = get_logger(name=__name__)


class ConnectionProcessor:
    """Process connection-level events and manage state transitions."""

    def __init__(self, *, is_client: bool, config: ClientConfig | ServerConfig, connection_id: ConnectionId) -> None:
        """Initialize the connection processor."""
        self._is_client = is_client
        self._config = config
        self._connection_id = connection_id

    def handle_cleanup_resources(self, *, event: InternalCleanupResources, state: ProtocolState) -> list[Effect]:
        """Handle the InternalCleanupResources event."""
        effects: list[Effect] = []

        closed_session_ids = {sid for sid, sdata in state.sessions.items() if sdata.state == SessionState.CLOSED}
        closed_stream_ids = {stid for stid, stdata in state.streams.items() if stdata.state == StreamState.CLOSED}

        for sid in closed_session_ids:
            logger.debug("Cleaning up closed session %s from state", sid)
            session_data = state.sessions.pop(sid, None)
            if session_data:
                control_stream_id = getattr(session_data, "control_stream_id", None)
                if control_stream_id is not None:
                    state.stream_to_session_map.pop(control_stream_id, None)
                    effects.append(CleanupH3Stream(stream_id=control_stream_id))

                assoc_streams = {stid for stid, st_sid in state.stream_to_session_map.items() if st_sid == sid}
                for stid in assoc_streams:
                    if stid in state.streams:
                        state.streams.pop(stid, None)
                    state.stream_to_session_map.pop(stid, None)
                    effects.append(CleanupH3Stream(stream_id=stid))

        for stid in closed_stream_ids:
            if stid in state.streams:
                logger.debug("Cleaning up closed stream %d from state", stid)
                state.streams.pop(stid, None)
                state.stream_to_session_map.pop(stid, None)
                effects.append(CleanupH3Stream(stream_id=stid))

        return effects

    def handle_connection_close(self, *, event: ConnectionClose, state: ProtocolState) -> list[Effect]:
        """Handle the ConnectionClose event."""
        effects: list[Effect] = []
        if state.connection_state not in (ConnectionState.CLOSED, ConnectionState.CLOSING):
            state.connection_state = ConnectionState.CLOSING
            state.closed_at = get_timestamp()
            effects.append(CloseQuicConnection(error_code=event.error_code, reason=event.reason))
        effects.append(CompleteUserFuture(future=event.future))
        return effects

    def handle_connection_terminated(
        self, *, event: TransportConnectionTerminated, state: ProtocolState
    ) -> list[Effect]:
        """Handle the TransportConnectionTerminated event."""
        if state.connection_state == ConnectionState.CLOSED:
            return []

        state.connection_state = ConnectionState.CLOSED
        state.closed_at = get_timestamp()

        effects: list[Effect] = []
        error = ConnectionError(message=f"Connection terminated: {event.reason_phrase}", error_code=event.error_code)

        pending_futures = state.pending_create_session_futures
        for stream_id, fut in list(pending_futures.items()):
            if not fut.done():
                effects.append(FailUserFuture(future=fut, exception=error))
            pending_futures.pop(stream_id, None)

        for stream_data in state.streams.values():
            while stream_data.pending_read_requests:
                read_fut = stream_data.pending_read_requests.popleft()
                if not read_fut.done():
                    effects.append(FailUserFuture(future=read_fut, exception=error))
            while stream_data.write_buffer:
                _data, write_fut, _end = stream_data.write_buffer.popleft()
                if not write_fut.done():
                    effects.append(FailUserFuture(future=write_fut, exception=error))

        effects.append(
            EmitConnectionEvent(
                event_type=EventType.CONNECTION_CLOSED,
                data={
                    "connection_id": self._connection_id,
                    "reason": event.reason_phrase,
                    "error_code": event.error_code,
                },
            )
        )

        return effects

    def handle_create_session(self, *, event: UserCreateSession, state: ProtocolState) -> list[Effect]:
        """Handle the UserCreateSession event (client-only)."""
        if not self._is_client:
            return [
                FailUserFuture(
                    future=event.future,
                    exception=ProtocolError(message="Server cannot create sessions using this method"),
                )
            ]

        if state.connection_state != ConnectionState.CONNECTED:
            return [
                FailUserFuture(
                    future=event.future,
                    exception=ConnectionError(
                        message=f"Cannot create session, connection state is {state.connection_state}"
                    ),
                )
            ]

        session_id = generate_session_id()
        now = get_timestamp()

        session_data = SessionStateData(
            session_id=session_id,
            control_stream_id=-1,
            state=SessionState.CONNECTING,
            path=event.path,
            headers=event.headers,
            created_at=now,
            local_max_data=self._config.initial_max_data,
            peer_max_data=state.peer_initial_max_data,
            local_max_streams_bidi=self._config.initial_max_streams_bidi,
            peer_max_streams_bidi=state.peer_initial_max_streams_bidi,
            local_max_streams_uni=self._config.initial_max_streams_uni,
            peer_max_streams_uni=state.peer_initial_max_streams_uni,
        )
        state.sessions[session_id] = session_data

        return [
            CreateH3Session(session_id=session_id, path=event.path, headers=event.headers, create_future=event.future)
        ]

    def handle_get_connection_diagnostics(
        self, *, event: UserGetConnectionDiagnostics, state: ProtocolState
    ) -> list[Effect]:
        """Handle the UserGetConnectionDiagnostics event."""
        diagnostics_data = {
            "connection_id": self._connection_id,
            "state": state.connection_state,
            "is_client": state.is_client,
            "connected_at": state.connected_at,
            "closed_at": state.closed_at,
            "max_datagram_size": state.max_datagram_size,
            "remote_max_datagram_frame_size": state.remote_max_datagram_frame_size,
            "session_count": len(state.sessions),
            "stream_count": len(state.streams),
        }
        return [CompleteUserFuture(future=event.future, value=diagnostics_data)]

    def handle_goaway_received(self, *, event: GoawayReceived, state: ProtocolState) -> list[Effect]:
        """Handle the H3 GOAWAY signal by draining all active sessions."""
        effects: list[Effect] = []

        if state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
            state.connection_state = ConnectionState.CLOSING
            state.closed_at = get_timestamp()

        for session_id, session_data in state.sessions.items():
            if session_data.state == SessionState.CONNECTED:
                session_data.state = SessionState.DRAINING
                effects.append(
                    SendH3Capsule(
                        stream_id=session_data.control_stream_id,
                        capsule_type=constants.DRAIN_WEBTRANSPORT_SESSION_TYPE,
                        capsule_data=b"",
                    )
                )
                effects.append(
                    EmitSessionEvent(
                        session_id=session_id, event_type=EventType.SESSION_DRAINING, data={"session_id": session_id}
                    )
                )
        return effects

    def handle_graceful_close(self, *, event: UserConnectionGracefulClose, state: ProtocolState) -> list[Effect]:
        """Handle the user request for a graceful H3 GOAWAY shutdown."""
        effects: list[Effect] = []
        if state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
            state.connection_state = ConnectionState.CLOSING
            state.closed_at = get_timestamp()
            effects.append(SendH3Goaway())
        effects.append(CompleteUserFuture(future=event.future))
        return effects

    def handle_headers_received(self, *, event: HeadersReceived, state: ProtocolState) -> list[Effect]:
        """Handle the HeadersReceived event."""
        effects: list[Effect] = []
        now = get_timestamp()
        stream_id = event.stream_id

        if self._is_client:
            session_id = state.stream_to_session_map.get(stream_id)
            if not session_id:
                logger.warning("Received headers on unknown client stream %d", stream_id)
                return []

            session_data = state.sessions.get(session_id)
            if not session_data or session_data.state != SessionState.CONNECTING:
                logger.warning("Received headers for non-connecting session %s on stream %d", session_id, stream_id)
                return []

            pending_futures = state.pending_create_session_futures
            create_future = pending_futures.pop(stream_id, None)

            if event.headers.get(":status") == "200":
                session_data.state = SessionState.CONNECTED
                session_data.ready_at = now
                effects.append(
                    EmitSessionEvent(
                        session_id=session_id,
                        event_type=EventType.SESSION_READY,
                        data={"session_id": session_id, "ready_at": now},
                    )
                )
                if create_future and not create_future.done():
                    effects.append(CompleteUserFuture(future=create_future, value=session_id))
            else:
                status = event.headers.get(":status", "Unknown")
                reason = f"Session creation failed with status {status}"
                session_data.state = SessionState.CLOSED
                session_data.closed_at = now
                session_data.close_reason = reason
                effects.append(
                    EmitSessionEvent(
                        session_id=session_id,
                        event_type=EventType.SESSION_CLOSED,
                        data={"session_id": session_id, "error_code": ErrorCodes.H3_REQUEST_REJECTED, "reason": reason},
                    )
                )
                if create_future and not create_future.done():
                    error = ConnectionError(message=reason, error_code=ErrorCodes.H3_REQUEST_REJECTED)
                    effects.append(FailUserFuture(future=create_future, exception=error))
                state.sessions.pop(session_id, None)
                state.stream_to_session_map.pop(stream_id, None)

        else:
            if state.connection_state != ConnectionState.CONNECTED:
                logger.debug(
                    "Rejecting new session on stream %d: connection state is %s", stream_id, state.connection_state
                )
                effects.append(SendH3Headers(stream_id=stream_id, status=429))
                return effects

            if event.headers.get(":method") != "CONNECT" or event.headers.get(":protocol") != "webtransport":
                logger.debug("Rejecting non-WebTransport request on stream %d", stream_id)
                effects.append(SendH3Headers(stream_id=stream_id, status=400))
                return effects

            max_sess = getattr(self._config, "max_sessions", 0)
            if max_sess > 0 and len(state.sessions) >= max_sess:
                logger.warning("Session limit (%d) reached, rejecting new session on stream %d", max_sess, stream_id)
                effects.append(SendH3Headers(stream_id=stream_id, status=429))
                return effects

            session_id = generate_session_id()
            path = event.headers.get(":path", "/")

            session_data = SessionStateData(
                session_id=session_id,
                control_stream_id=stream_id,
                state=SessionState.CONNECTING,
                path=path,
                headers=event.headers,
                created_at=now,
                local_max_data=self._config.initial_max_data,
                peer_max_data=state.peer_initial_max_data,
                local_max_streams_bidi=self._config.initial_max_streams_bidi,
                peer_max_streams_bidi=state.peer_initial_max_streams_bidi,
                local_max_streams_uni=self._config.initial_max_streams_uni,
                peer_max_streams_uni=state.peer_initial_max_streams_uni,
            )
            state.sessions[session_id] = session_data
            state.stream_to_session_map[stream_id] = session_id

            effects.append(
                EmitSessionEvent(
                    session_id=session_id,
                    event_type=EventType.SESSION_REQUEST,
                    data={
                        "session_id": session_id,
                        "control_stream_id": stream_id,
                        "path": path,
                        "headers": event.headers,
                    },
                )
            )

        return effects

    def handle_transport_parameters_received(
        self, *, event: TransportQuicParametersReceived, state: ProtocolState
    ) -> list[Effect]:
        """Handle the TransportQuicParametersReceived event."""
        logger.debug(
            "Received transport parameters: remote_max_datagram_frame_size=%d", event.remote_max_datagram_frame_size
        )
        state.remote_max_datagram_frame_size = event.remote_max_datagram_frame_size
        return []



"""High-level Remote Procedure Call (RPC) pattern."""

from .exceptions import InvalidParamsError, MethodNotFoundError, RpcError, RpcErrorCode, RpcTimeoutError
from .manager import RpcManager
from .protocol import RpcErrorResponse, RpcRequest, RpcSuccessResponse

__all__: list[str] = [
    "InvalidParamsError",
    "MethodNotFoundError",
    "RpcError",
    "RpcErrorCode",
    "RpcErrorResponse",
    "RpcManager",
    "RpcRequest",
    "RpcSuccessResponse",
    "RpcTimeoutError",
]



"""Core WebTransport connection object representing a QUIC connection."""

from __future__ import annotations

import asyncio
import uuid
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Self, TypeAlias, cast

from pywebtransport._protocol.events import (
    ConnectionClose,
    Effect,
    EmitConnectionEvent,
    EmitSessionEvent,
    EmitStreamEvent,
    UserConnectionGracefulClose,
    UserCreateSession,
    UserEvent,
    UserGetConnectionDiagnostics,
)
from pywebtransport._protocol.webtransport_engine import WebTransportEngine
from pywebtransport.constants import ErrorCodes
from pywebtransport.events import EventEmitter
from pywebtransport.exceptions import ConnectionError, SessionError
from pywebtransport.session.session import WebTransportSession
from pywebtransport.stream.stream import WebTransportReceiveStream, WebTransportSendStream, WebTransportStream
from pywebtransport.types import (
    Address,
    ConnectionId,
    ConnectionState,
    EventType,
    Headers,
    SessionId,
    StreamDirection,
    StreamId,
)
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport.client._adapter import WebTransportClientProtocol
    from pywebtransport.config import ClientConfig, ServerConfig
    from pywebtransport.server._adapter import WebTransportServerProtocol

    AdapterProtocol: TypeAlias = WebTransportServerProtocol | WebTransportClientProtocol
    StreamHandle: TypeAlias = WebTransportStream | WebTransportReceiveStream | WebTransportSendStream

__all__: list[str] = ["ConnectionDiagnostics", "WebTransportConnection"]

logger = get_logger(name=__name__)


@dataclass(kw_only=True)
class ConnectionDiagnostics:
    """A snapshot of connection diagnostics."""

    connection_id: ConnectionId
    state: ConnectionState
    is_client: bool
    connected_at: float | None
    closed_at: float | None
    max_datagram_size: int
    remote_max_datagram_frame_size: int
    session_count: int
    stream_count: int
    active_session_handles: int
    active_stream_handles: int


class WebTransportConnection:
    """A high-level handle for a WebTransport connection over QUIC."""

    def __init__(
        self,
        *,
        config: ClientConfig | ServerConfig,
        protocol: AdapterProtocol,
        transport: asyncio.DatagramTransport,
        is_client: bool,
    ) -> None:
        """Initialize the WebTransport connection."""
        self._config = config
        self._protocol = protocol
        self._transport = transport
        self._is_client = is_client
        self._connection_id: ConnectionId = str(uuid.uuid4())
        self.events = EventEmitter()

        self._engine = WebTransportEngine(
            connection_id=self._connection_id,
            config=config,
            is_client=is_client,
            protocol_handler=protocol,
            owner_notify_callback=self._notify_owner,
        )

        self._session_handles: dict[SessionId, WebTransportSession] = {}
        self._stream_handles: dict[StreamId, StreamHandle] = {}

        logger.debug("WebTransportConnection %s initialized.", self.connection_id)

    @property
    def config(self) -> ClientConfig | ServerConfig:
        """Get the configuration associated with this connection."""
        return self._config

    @property
    def connection_id(self) -> ConnectionId:
        """Get the unique identifier for this connection."""
        return self._connection_id

    @property
    def is_client(self) -> bool:
        """Return True if this is a client-side connection."""
        return self._is_client

    @property
    def is_closed(self) -> bool:
        """Return True if the connection is closed."""
        return self.state == ConnectionState.CLOSED

    @property
    def is_closing(self) -> bool:
        """Return True if the connection is closing."""
        return self.state == ConnectionState.CLOSING

    @property
    def is_connected(self) -> bool:
        """Return True if the connection is established."""
        return self.state == ConnectionState.CONNECTED

    @property
    def local_address(self) -> Address | None:
        """Get the local address of the connection."""
        if hasattr(self, "_transport") and self._transport:
            return cast(Address | None, self._transport.get_extra_info("sockname"))
        return None

    @property
    def remote_address(self) -> Address | None:
        """Get the remote address of the connection."""
        if hasattr(self, "_transport") and self._transport:
            return cast(Address | None, self._transport.get_extra_info("peername"))
        return None

    @property
    def state(self) -> ConnectionState:
        """Get the current state of the connection."""
        return self._get_engine_state()

    async def __aenter__(self) -> Self:
        """Enter the async context manager."""
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context manager."""
        await self.close()

    async def close(self, *, error_code: int = ErrorCodes.NO_ERROR, reason: str = "Closed by application") -> None:
        """Immediately close the WebTransport connection."""
        logger.info("Closing connection %s...", self.connection_id)

        fut = asyncio.get_running_loop().create_future()
        event = ConnectionClose(error_code=error_code, reason=reason, future=fut)

        await self._send_event_to_engine(event=event)

        try:
            async with asyncio.timeout(delay=5.0):
                await fut
        except asyncio.TimeoutError:
            fut.cancel()
        except Exception:
            pass

        if hasattr(self, "_engine"):
            await self._engine.stop_driver_loop()

        if self.is_client:
            if hasattr(self, "_transport") and self._transport and not self._transport.is_closing():
                logger.debug("Closing underlying transport for connection %s", self.connection_id)
                self._transport.close()

        self._session_handles.clear()
        self._stream_handles.clear()

        if hasattr(self, "_engine") and hasattr(self._engine, "_state"):
            self._engine._state.connection_state = ConnectionState.CLOSED
        logger.info("Connection %s close process finished.", self.connection_id)

    async def create_session(self, *, path: str, headers: Headers | None = None) -> WebTransportSession:
        """Create a new WebTransport session."""
        if not self.is_client:
            raise ConnectionError("Sessions can only be created by the client.")
        if not hasattr(self, "_engine"):
            raise ConnectionError("Engine not available for creating session.")

        fut = asyncio.get_running_loop().create_future()
        event = UserCreateSession(path=path, headers=headers or {}, future=fut)
        await self._send_event_to_engine(event=event)

        session_id: SessionId = await fut

        session_handle = self._session_handles.get(session_id)
        if not session_handle:
            logger.error("Internal error: Session handle %s missing after successful creation effect.", session_id)
            raise SessionError(f"Internal error creating session handle for {session_id}")

        return session_handle

    async def graceful_shutdown(self) -> None:
        """Gracefully shut down the connection."""
        logger.info("Initiating graceful shutdown for connection %s...", self.connection_id)
        fut = asyncio.get_running_loop().create_future()
        event = UserConnectionGracefulClose(future=fut)

        await self._send_event_to_engine(event=event)

        try:
            async with asyncio.timeout(delay=5.0):
                await fut
        except asyncio.TimeoutError:
            fut.cancel()
            logger.warning("Timeout waiting for graceful shutdown GOAWAY confirmation.")
        except Exception as e:
            logger.warning("Error during graceful shutdown: %s", e)

        await self.close(reason="Graceful shutdown complete")

    async def initialize(self) -> None:
        """Start the internal engine loop."""
        if hasattr(self, "_engine") and self._engine._driver_task is None:
            self._engine.start_driver_loop()
            if self._engine._event_queue is not None:
                self._protocol.set_engine_queue(engine_queue=self._engine._event_queue)
            else:
                logger.critical("Engine failed to create event queue during initialization for %s", self.connection_id)
                await self.close()
        elif not hasattr(self, "_engine"):
            logger.warning("Attempted to initialize connection %s without an engine.", self.connection_id)

    async def diagnostics(self) -> ConnectionDiagnostics:
        """Get diagnostic information about the connection."""
        if not hasattr(self, "_engine"):
            raise ConnectionError("Engine not available.")

        fut = asyncio.get_running_loop().create_future()
        event = UserGetConnectionDiagnostics(future=fut)
        await self._send_event_to_engine(event=event)

        diag_data: dict[str, Any] = await fut
        diag_data["active_session_handles"] = len(self._session_handles)
        diag_data["active_stream_handles"] = len(self._stream_handles)
        return ConnectionDiagnostics(**diag_data)

    def get_all_sessions(self) -> list[WebTransportSession]:
        """Get a list of all active session handles."""
        return list(self._session_handles.values())

    def _get_engine_state(self) -> ConnectionState:
        """Safely get the current engine state for internal logic."""
        if hasattr(self, "_engine") and hasattr(self._engine, "_state"):
            return self._engine._state.connection_state
        return ConnectionState.CLOSED

    def _notify_owner(self, effect: Effect) -> None:
        """Handle events emitted from the protocol engine."""
        try:
            match effect:
                case EmitConnectionEvent(event_type=et, data=d):
                    if "connection" not in d:
                        d["connection"] = self
                    self.events.emit_nowait(event_type=et, data=d)

                case EmitSessionEvent(session_id=sid, event_type=et, data=d):
                    create_handle_event = (not self.is_client and et == EventType.SESSION_REQUEST) or (
                        self.is_client and et == EventType.SESSION_READY
                    )

                    session_handle = self._session_handles.get(sid)

                    if not session_handle and create_handle_event:
                        if hasattr(self, "_engine") and hasattr(self._engine, "_state"):
                            session_data = self._engine._state.sessions.get(sid)
                            if session_data:
                                session = WebTransportSession(
                                    connection=self,
                                    session_id=sid,
                                    path=session_data.path,
                                    headers=session_data.headers,
                                    control_stream_id=session_data.control_stream_id,
                                )
                                self._session_handles[sid] = session
                                session_handle = session
                                logger.debug("Created session handle for %s", sid)
                            else:
                                logger.error("Engine state for session %s missing during handle creation.", sid)
                        else:
                            logger.error("Engine or state missing during session handle creation for %s.", sid)

                    if session_handle:
                        if "session" not in d:
                            d["session"] = session_handle

                        if et in (EventType.SESSION_DATA_BLOCKED, EventType.SESSION_STREAMS_BLOCKED):
                            if session_handle.events.listener_count(event_type=et) == 0:
                                logger.warning("Session %s received unhandled event '%s'.", sid, et.value)

                        session_handle.events.emit_nowait(event_type=et, data=d)

                        if et == EventType.SESSION_REQUEST:
                            self.events.emit_nowait(event_type=et, data=d)
                    elif not create_handle_event:
                        logger.warning("No session handle found for event %s on session %s", et, sid)

                    if et == EventType.SESSION_CLOSED:
                        removed = self._session_handles.pop(sid, None)
                        if removed:
                            logger.debug("Removed session handle for closed session %s", sid)
                            asyncio.create_task(coro=removed.events.close())

                case EmitStreamEvent(stream_id=stid, event_type=et, data=d):
                    stream_handle = self._stream_handles.get(stid)
                    stream_session_handle: WebTransportSession | None = None

                    if not stream_handle and et == EventType.STREAM_OPENED:
                        if hasattr(self, "_engine") and hasattr(self._engine, "_state"):
                            stream_data = self._engine._state.streams.get(stid)
                            stream_session_handle = (
                                self._session_handles.get(stream_data.session_id) if stream_data else None
                            )
                            if stream_data and stream_session_handle:
                                handle_class: type[StreamHandle]
                                match stream_data.direction:
                                    case StreamDirection.BIDIRECTIONAL:
                                        handle_class = WebTransportStream
                                    case StreamDirection.SEND_ONLY:
                                        handle_class = WebTransportSendStream
                                    case StreamDirection.RECEIVE_ONLY:
                                        handle_class = WebTransportReceiveStream

                                new_stream_handle = handle_class(session=stream_session_handle, stream_id=stid)
                                self._stream_handles[stid] = new_stream_handle
                                stream_handle = new_stream_handle
                                logger.debug("Created stream handle for %d (%s)", stid, stream_data.direction)
                            else:
                                logger.error(
                                    "Stream data (%s) or session handle (%s) missing for stream %d.",
                                    bool(stream_data),
                                    bool(stream_session_handle),
                                    stid,
                                )
                        else:
                            logger.error("Engine or state missing during stream handle creation for %d.", stid)

                    if stream_handle:
                        if "stream" not in d:
                            d["stream"] = stream_handle
                        stream_handle.events.emit_nowait(event_type=et, data=d)

                        if et == EventType.STREAM_OPENED:
                            if stream_session_handle is None:
                                stream_data = self._engine._state.streams.get(stid)
                                if stream_data:
                                    stream_session_handle = self._session_handles.get(stream_data.session_id)

                            if stream_session_handle:
                                stream_session_handle._add_stream_handle(stream=stream_handle, event_data=d)
                            else:
                                logger.error("No session handle found to propagate STREAM_OPENED for stream %d", stid)

                    elif et != EventType.STREAM_OPENED:
                        logger.warning("No stream handle found for event %s on stream %d", et, stid)

                    if et == EventType.STREAM_CLOSED:
                        removed_stream = self._stream_handles.pop(stid, None)
                        if removed_stream:
                            logger.debug("Removed stream handle for closed stream %d", stid)
                            asyncio.create_task(coro=removed_stream.events.close())

        except Exception as e:
            logger.error("Error during owner notification callback: %s", e, exc_info=True)

    async def _send_event_to_engine(self, *, event: UserEvent) -> None:
        """Send a UserEvent to the protocol engine."""
        if not hasattr(self, "_engine"):
            if not event.future.done():
                exc = ConnectionError(f"Connection {self.connection_id} engine is missing.")
                try:
                    event.future.set_exception(exc)
                except asyncio.InvalidStateError:
                    pass
            return

        await self._engine.put_event(event=event)

    def __repr__(self) -> str:
        """Provide a developer-friendly representation."""
        current_state = self._get_engine_state()
        return f"<WebTransportConnection id={self.connection_id} state={current_state} client={self.is_client}>"



"""High-level abstraction for a WebTransport session."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Self

from pywebtransport._protocol.events import (
    UserCloseSession,
    UserCreateStream,
    UserEvent,
    UserGetSessionDiagnostics,
    UserGrantDataCredit,
    UserGrantStreamsCredit,
    UserSendDatagram,
)
from pywebtransport.events import EventEmitter
from pywebtransport.exceptions import ConnectionError, SessionError, StreamError, TimeoutError
from pywebtransport.stream.stream import WebTransportSendStream, WebTransportStream
from pywebtransport.types import Data, EventType, Headers, SessionId, SessionState, StreamId
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport._protocol.state import SessionStateData
    from pywebtransport.connection.connection import WebTransportConnection
    from pywebtransport.stream.stream import StreamType


__all__: list[str] = ["SessionDiagnostics", "WebTransportSession"]

logger = get_logger(name=__name__)


@dataclass(kw_only=True)
class SessionDiagnostics:
    """A snapshot of session diagnostics."""

    session_id: SessionId
    control_stream_id: StreamId
    state: SessionState
    path: str
    headers: Headers
    created_at: float
    local_max_data: int
    local_data_sent: int
    peer_max_data: int
    peer_data_sent: int
    local_max_streams_bidi: int
    local_streams_bidi_opened: int
    peer_max_streams_bidi: int
    peer_streams_bidi_opened: int
    local_max_streams_uni: int
    local_streams_uni_opened: int
    peer_max_streams_uni: int
    peer_streams_uni_opened: int
    pending_bidi_stream_futures: list[Any]
    pending_uni_stream_futures: list[Any]
    datagrams_sent: int
    datagram_bytes_sent: int
    datagrams_received: int
    datagram_bytes_received: int
    close_code: int | None
    close_reason: str | None
    closed_at: float | None
    ready_at: float | None


class WebTransportSession:
    """A high-level handle for a WebTransport session."""

    _connection: WebTransportConnection
    _session_id: SessionId
    _control_stream_id: StreamId | None
    _path: str
    _headers: Headers
    events: EventEmitter

    def __init__(
        self,
        *,
        connection: WebTransportConnection,
        session_id: SessionId,
        path: str,
        headers: Headers,
        control_stream_id: StreamId | None,
    ) -> None:
        """Initialize the WebTransportSession handle."""
        self._connection = connection
        self._session_id = session_id
        self._control_stream_id = control_stream_id
        self._path = path
        self._headers = headers
        self.events = EventEmitter()

        logger.debug("WebTransportSession handle created for session %s", self._session_id)

    @property
    def headers(self) -> Headers:
        """Get the initial request headers for this session."""
        return self._headers.copy()

    @property
    def is_closed(self) -> bool:
        """Return True if the session is closed."""
        session_data = self._get_engine_state()
        return not session_data or session_data.state == SessionState.CLOSED

    @property
    def path(self) -> str:
        """Get the request path associated with this session."""
        return self._path

    @property
    def session_id(self) -> SessionId:
        """Get the unique identifier for this session."""
        return self._session_id

    @property
    def state(self) -> SessionState:
        """Get the current state of the session."""
        session_data = self._get_engine_state()
        return session_data.state if session_data else SessionState.CLOSED

    async def __aenter__(self) -> Self:
        """Enter async context."""
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit async context, closing the session."""
        await self.close()

    async def close(self, *, error_code: int = 0, reason: str | None = None, close_connection: bool = False) -> None:
        """Close the WebTransport session."""
        logger.info("Closing session %s: code=%#x reason='%s'", self.session_id, error_code, reason or "")
        fut = asyncio.get_running_loop().create_future()
        event = UserCloseSession(session_id=self.session_id, error_code=error_code, reason=reason, future=fut)
        await self._connection._send_event_to_engine(event=event)
        try:
            await fut
        except SessionError as e:
            logger.warning("Error initiating session close for %s: %s", self.session_id, e)

        if close_connection:
            logger.debug("Closing parent connection for session %s", self.session_id)
            await self._connection.close()

    async def create_bidirectional_stream(self) -> WebTransportStream:
        """Create a new bidirectional WebTransport stream."""
        fut = asyncio.get_running_loop().create_future()
        event = UserCreateStream(session_id=self.session_id, is_unidirectional=False, future=fut)
        await self._connection._send_event_to_engine(event=event)

        try:
            timeout = self._connection.config.stream_creation_timeout
            async with asyncio.timeout(delay=timeout):
                stream_id: StreamId = await fut
        except asyncio.TimeoutError:
            fut.cancel()
            raise TimeoutError(
                f"Session {self.session_id} timed out creating bidirectional stream after {timeout}s"
            ) from None
        except Exception as e:
            raise e

        stream_handle = self._connection._stream_handles.get(stream_id)
        if not isinstance(stream_handle, WebTransportStream):
            logger.error("Internal error: Stream handle %d missing or wrong type after creation", stream_id)
            raise StreamError(f"Internal error creating stream handle for {stream_id}")

        return stream_handle

    async def create_unidirectional_stream(self) -> WebTransportSendStream:
        """Create a new unidirectional (send-only) WebTransport stream."""
        fut = asyncio.get_running_loop().create_future()
        event = UserCreateStream(session_id=self.session_id, is_unidirectional=True, future=fut)
        await self._connection._send_event_to_engine(event=event)

        try:
            timeout = self._connection.config.stream_creation_timeout
            async with asyncio.timeout(delay=timeout):
                stream_id: StreamId = await fut
        except asyncio.TimeoutError:
            fut.cancel()
            raise TimeoutError(
                f"Session {self.session_id} timed out creating unidirectional stream after {timeout}s"
            ) from None
        except Exception as e:
            raise e

        stream_handle = self._connection._stream_handles.get(stream_id)
        if not isinstance(stream_handle, WebTransportSendStream) or isinstance(stream_handle, WebTransportStream):
            logger.error("Internal error: Stream handle %d missing or wrong type after creation", stream_id)
            raise StreamError(f"Internal error creating stream handle for {stream_id}")

        return stream_handle

    async def diagnostics(self) -> SessionDiagnostics:
        """Get diagnostic information about the session."""
        fut = asyncio.get_running_loop().create_future()
        event = UserGetSessionDiagnostics(session_id=self.session_id, future=fut)
        try:
            await self._connection._send_event_to_engine(event=event)
        except ConnectionError as e:
            raise SessionError(f"Connection is closed, cannot get diagnostics: {e}") from e

        diag_data: dict[str, Any] = await fut
        return SessionDiagnostics(**diag_data)

    async def grant_data_credit(self, *, max_data: int) -> None:
        """Manually grant data flow control credit to the peer."""
        fut = asyncio.get_running_loop().create_future()
        event = UserGrantDataCredit(session_id=self.session_id, max_data=max_data, future=fut)
        await self._connection._send_event_to_engine(event=event)
        await fut

    async def grant_streams_credit(self, *, max_streams: int, is_unidirectional: bool) -> None:
        """Manually grant stream flow control credit to the peer."""
        fut = asyncio.get_running_loop().create_future()
        event = UserGrantStreamsCredit(
            session_id=self.session_id, max_streams=max_streams, is_unidirectional=is_unidirectional, future=fut
        )
        await self._connection._send_event_to_engine(event=event)
        await fut

    async def send_datagram(self, *, data: Data) -> None:
        """Send an unreliable datagram."""
        fut = asyncio.get_running_loop().create_future()
        event = UserSendDatagram(session_id=self.session_id, data=data, future=fut)
        await self._connection._send_event_to_engine(event=event)
        await fut

    def _add_stream_handle(self, *, stream: StreamType, event_data: dict[str, Any]) -> None:
        """Register an incoming stream and re-emit the STREAM_OPENED event."""
        logger.debug("Session %s re-emitting STREAM_OPENED for stream %s", self.session_id, stream.stream_id)

        event_payload = event_data.copy()
        event_payload["stream"] = stream

        self.events.emit_nowait(event_type=EventType.STREAM_OPENED, data=event_payload)

    def _get_engine_state(self) -> SessionStateData | None:
        """Safely get the current engine state for internal logic."""
        if (conn := self._connection) and hasattr(conn, "_engine"):
            if hasattr(conn._engine, "_state"):
                return conn._engine._state.sessions.get(self._session_id)
        return None

    async def _send_event_to_engine(self, *, event: UserEvent) -> None:
        """Send a UserEvent to the protocol engine (via the connection)."""
        await self._connection._send_event_to_engine(event=event)

    def __repr__(self) -> str:
        """Provide a developer-friendly representation."""
        session_data = self._get_engine_state()
        current_state = session_data.state if session_data else SessionState.CLOSED
        return f"<WebTransportSession id={self.session_id} state={current_state}>"



"""Utility for monitoring server health."""

from __future__ import annotations

from dataclasses import asdict
from typing import Any

from pywebtransport.monitor._base import _BaseMonitor
from pywebtransport.server.server import ServerDiagnostics, WebTransportServer
from pywebtransport.types import ConnectionState
from pywebtransport.utils import get_logger, get_timestamp

__all__: list[str] = ["ServerMonitor"]

logger = get_logger(name=__name__)


class ServerMonitor(_BaseMonitor[WebTransportServer]):
    """Monitor server performance and health via an async context."""

    def __init__(self, server: WebTransportServer, *, monitoring_interval: float = 30.0) -> None:
        """Initialize the server monitor."""
        super().__init__(target=server, monitoring_interval=monitoring_interval)

    def get_current_metrics(self) -> dict[str, Any] | None:
        """Get the latest collected metrics."""
        return self._metrics_history[-1] if self._metrics_history else None

    def get_health_status(self) -> dict[str, Any]:
        """Get the current server health status based on the latest metrics."""
        metrics = self.get_current_metrics()
        if metrics is None:
            return {"status": "unknown", "reason": "No metrics collected yet."}

        if not self._target.is_serving:
            return {"status": "unhealthy", "reason": "Server is not serving."}

        stats = metrics.get("stats", {})
        connection_states = metrics.get("connection_states", {})

        accepted = stats.get("connections_accepted", 0)
        rejected = stats.get("connections_rejected", 0)
        total_attempts = accepted + rejected

        if total_attempts > 10:
            success_rate = accepted / total_attempts if total_attempts > 0 else 1.0
            if success_rate < 0.9:
                return {"status": "degraded", "reason": f"Low connection success rate: {success_rate:.2%}"}

        active_connections = connection_states.get(ConnectionState.CONNECTED.value, 0)
        if active_connections > 0:
            return {"status": "healthy", "reason": "Server is operating normally."}

        return {"status": "idle", "reason": "Server is running but has no active connections."}

    def _check_for_alerts(self) -> None:
        """Analyze the latest metrics and generate alerts if thresholds are breached."""
        try:
            health = self.get_health_status()
            if health["status"] in ("unhealthy", "degraded"):
                if not self._alerts or self._alerts[-1].get("reason") != health["reason"]:
                    alert = {"timestamp": get_timestamp(), "status": health["status"], "reason": health["reason"]}
                    self._alerts.append(alert)
                    logger.warning("Health Alert: %s - %s", health["status"], health["reason"])
        except Exception as e:
            logger.error("Alert check failed: %s", e, exc_info=True)

    async def _collect_metrics(self) -> None:
        """Collect a snapshot of the server's current statistics."""
        try:
            timestamp = get_timestamp()
            diagnostics_obj: ServerDiagnostics = await self._target.diagnostics()
            metrics = {"timestamp": timestamp, **asdict(diagnostics_obj)}
            self._metrics_history.append(metrics)
        except Exception as e:
            logger.error("Metrics collection failed: %s", e, exc_info=True)



"""Internal protocol engine for driving the WebTransport state machine."""

from __future__ import annotations

import asyncio
import weakref
from collections import deque
from collections.abc import Callable
from typing import TYPE_CHECKING, TypeAlias

from aioquic.buffer import encode_uint_var

from pywebtransport import constants
from pywebtransport._protocol.connection_processor import ConnectionProcessor
from pywebtransport._protocol.events import (
    CapsuleReceived,
    CleanupH3Stream,
    CloseQuicConnection,
    CompleteUserFuture,
    ConnectionClose,
    ConnectStreamClosed,
    CreateH3Session,
    CreateQuicStream,
    DatagramReceived,
    Effect,
    EmitConnectionEvent,
    EmitSessionEvent,
    EmitStreamEvent,
    FailUserFuture,
    GoawayReceived,
    HeadersReceived,
    InternalCleanupEarlyEvents,
    InternalCleanupResources,
    LogH3Frame,
    ProtocolEvent,
    RescheduleQuicTimer,
    ResetQuicStream,
    SendH3Capsule,
    SendH3Datagram,
    SendH3Goaway,
    SendH3Headers,
    SendQuicData,
    SendQuicDatagram,
    SettingsReceived,
    StopQuicStream,
    TransportConnectionTerminated,
    TransportDatagramFrameReceived,
    TransportHandshakeCompleted,
    TransportQuicParametersReceived,
    TransportQuicTimerFired,
    TransportStreamDataReceived,
    TransportStreamReset,
    TriggerQuicTimer,
    UserAcceptSession,
    UserCloseSession,
    UserConnectionGracefulClose,
    UserCreateSession,
    UserCreateStream,
    UserEvent,
    UserGetConnectionDiagnostics,
    UserGetSessionDiagnostics,
    UserGetStreamDiagnostics,
    UserGrantDataCredit,
    UserGrantStreamsCredit,
    UserRejectSession,
    UserResetStream,
    UserSendDatagram,
    UserSendStreamData,
    UserStopStream,
    UserStreamRead,
    WebTransportStreamDataReceived,
)
from pywebtransport._protocol.h3_engine import WebTransportH3Engine
from pywebtransport._protocol.session_processor import SessionProcessor
from pywebtransport._protocol.state import ProtocolState, StreamStateData
from pywebtransport._protocol.stream_processor import StreamProcessor
from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import ConnectionError, ProtocolError, SessionError
from pywebtransport.types import (
    ConnectionId,
    ConnectionState,
    EventType,
    Headers,
    SessionId,
    StreamDirection,
    StreamState,
)
from pywebtransport.utils import get_logger, get_timestamp

if TYPE_CHECKING:
    from pywebtransport.client._adapter import WebTransportClientProtocol
    from pywebtransport.config import ClientConfig, ServerConfig
    from pywebtransport.server._adapter import WebTransportServerProtocol

    AdapterProtocol: TypeAlias = WebTransportServerProtocol | WebTransportClientProtocol
    NotifyCallback: TypeAlias = Callable[[Effect], None]


__all__: list[str] = []

logger = get_logger(name=__name__)


class WebTransportEngine:
    """Orchestrates the unified protocol state machine."""

    def __init__(
        self,
        *,
        connection_id: ConnectionId,
        config: ClientConfig | ServerConfig,
        is_client: bool,
        protocol_handler: AdapterProtocol,
        owner_notify_callback: NotifyCallback,
    ) -> None:
        """Initialize the WebTransport engine."""
        self._connection_id = connection_id
        self._is_client = is_client
        self._protocol_handler = weakref.ref(protocol_handler)
        self._owner_notify_callback = owner_notify_callback
        self._event_queue: asyncio.Queue[ProtocolEvent] | None = None
        self._driver_task: asyncio.Task[None] | None = None
        self._early_event_cleanup_task: asyncio.Task[None] | None = None
        self._resource_gc_timer_task: asyncio.Task[None] | None = None

        self._pending_event_ttl = config.pending_event_ttl
        self._resource_cleanup_interval = config.resource_cleanup_interval
        self._next_early_event_cleanup_at: float = 0.0

        self._state = ProtocolState(
            is_client=is_client, connection_state=ConnectionState.IDLE, max_datagram_size=config.max_datagram_size
        )

        self._connection_processor = ConnectionProcessor(
            is_client=is_client, config=config, connection_id=connection_id
        )
        self._session_processor = SessionProcessor(is_client=is_client, config=config)
        self._stream_processor = StreamProcessor(is_client=is_client, config=config)
        self._h3_engine = WebTransportH3Engine(is_client=is_client, config=config)

        self._pending_user_actions: deque[UserEvent] = deque()

    async def put_event(self, *, event: ProtocolEvent) -> None:
        """Place a new event into the engine's processing queue."""
        if self._driver_task is not None and not self._driver_task.done() and self._event_queue is not None:
            await self._event_queue.put(event)
        else:
            if isinstance(event, UserEvent) and not event.future.done():
                error = ConnectionError("Engine is not running.")
                try:
                    event.future.set_exception(error)
                except asyncio.InvalidStateError:
                    pass
            logger.warning("Engine not running, discarding event: %s", type(event).__name__)

    def start_driver_loop(self) -> None:
        """Start the main event processing loop as a background task."""
        if self._state.connection_state == ConnectionState.IDLE:
            self._state.connection_state = ConnectionState.CONNECTING
        if self._event_queue is None:
            self._event_queue = asyncio.Queue()

        if self._driver_task is None:
            self._driver_task = asyncio.create_task(coro=self._driver_loop())
            logger.debug("WebTransportEngine driver loop started.")

        if self._early_event_cleanup_task is None and self._pending_event_ttl > 0:
            self._early_event_cleanup_task = asyncio.create_task(coro=self._early_event_cleanup_loop())

        if self._resource_gc_timer_task is None and self._resource_cleanup_interval > 0:
            self._resource_gc_timer_task = asyncio.create_task(coro=self._resource_gc_timer_loop())

    async def stop_driver_loop(self) -> None:
        """Stop the main event processing loop."""
        if self._early_event_cleanup_task and not self._early_event_cleanup_task.done():
            self._early_event_cleanup_task.cancel()
            try:
                await self._early_event_cleanup_task
            except asyncio.CancelledError:
                pass
        self._early_event_cleanup_task = None

        if self._resource_gc_timer_task and not self._resource_gc_timer_task.done():
            self._resource_gc_timer_task.cancel()
            try:
                await self._resource_gc_timer_task
            except asyncio.CancelledError:
                pass
        self._resource_gc_timer_task = None

        if self._driver_task and not self._driver_task.done():
            self._driver_task.cancel()
            try:
                await self._driver_task
            except asyncio.CancelledError:
                pass
            logger.debug("WebTransportEngine driver loop stopped.")
        self._driver_task = None

        if self._event_queue:
            while not self._event_queue.empty():
                event = self._event_queue.get_nowait()
                if isinstance(event, UserEvent) and not event.future.done():
                    error = ConnectionError("Engine stopped")
                    try:
                        event.future.set_exception(error)
                    except asyncio.InvalidStateError:
                        pass
                self._event_queue.task_done()

    def _check_client_connection_ready(self) -> tuple[list[Effect], bool]:
        """Check if the client connection is fully ready (QUIC + H3)."""
        if (
            self._state.connection_state == ConnectionState.CONNECTING
            and self._state.handshake_complete
            and self._state.peer_settings_received
        ):
            logger.debug("Client connection fully ready (QUIC + H3 SETTINGS).")
            self._state.connection_state = ConnectionState.CONNECTED
            self._state.connected_at = get_timestamp()
            effects: list[Effect] = [
                EmitConnectionEvent(
                    event_type=EventType.CONNECTION_ESTABLISHED, data={"connection_id": self._connection_id}
                )
            ]
            return effects, True

        return [], False

    async def _driver_loop(self) -> None:
        """The main event processing loop."""
        final_reason: str | None = "Driver loop ended unexpectedly"
        final_error_code: int = ErrorCodes.INTERNAL_ERROR
        last_effects: list[Effect] = []
        try:
            if self._event_queue is None:
                logger.critical("Driver loop started without an event queue.")
                return

            while self._state.connection_state != ConnectionState.CLOSED:
                event = await self._event_queue.get()
                current_effects: list[Effect] = []
                try:
                    current_effects = self._handle_event(event=event)
                    last_effects = current_effects

                except ProtocolError as pe:
                    logger.debug("Protocol error occurred: %s (Code: %#x)", pe.message, pe.error_code)
                    final_reason = f"Protocol error: {pe.message}"
                    final_error_code = pe.error_code
                    if self._state.connection_state != ConnectionState.CLOSING:
                        self._state.connection_state = ConnectionState.CLOSING
                        current_effects.append(CloseQuicConnection(error_code=final_error_code, reason=final_reason))

                except Exception as e:
                    logger.error("Unexpected internal error in event loop: %s", e, exc_info=True)
                    final_reason = f"Unexpected internal error: {e}"
                    final_error_code = ErrorCodes.INTERNAL_ERROR
                    if self._state.connection_state != ConnectionState.CLOSING:
                        self._state.connection_state = ConnectionState.CLOSING
                        current_effects.append(CloseQuicConnection(error_code=final_error_code, reason=final_reason))

                await self._execute_effects(effects=current_effects)

                self._event_queue.task_done()

            if self._state.connection_state == ConnectionState.CLOSED:
                for eff in last_effects:
                    if isinstance(eff, CloseQuicConnection):
                        final_reason = eff.reason
                        final_error_code = eff.error_code
                        break
                else:
                    if final_reason == "Driver loop ended unexpectedly":
                        final_reason = "Connection closed"
                        final_error_code = ErrorCodes.NO_ERROR

        except asyncio.CancelledError:
            logger.debug("Engine driver loop cancelled.")
            final_reason = "Driver loop cancelled"
            final_error_code = ErrorCodes.NO_ERROR
        except Exception as e:
            logger.critical("Fatal error in engine driver loop: %s", e, exc_info=True)
            final_reason = "Fatal engine error"
            final_error_code = ErrorCodes.INTERNAL_ERROR
            if self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                close_effects_on_error_fatal: list[Effect] = [
                    CloseQuicConnection(error_code=final_error_code, reason=final_reason)
                ]
                try:
                    await self._execute_effects(effects=close_effects_on_error_fatal)
                except Exception:
                    logger.error("Ignoring error during final close attempt in fatal handler.")
        finally:
            if self._early_event_cleanup_task and not self._early_event_cleanup_task.done():
                self._early_event_cleanup_task.cancel()
                try:
                    await self._early_event_cleanup_task
                except asyncio.CancelledError:
                    pass
            self._early_event_cleanup_task = None

            if self._resource_gc_timer_task and not self._resource_gc_timer_task.done():
                self._resource_gc_timer_task.cancel()
                try:
                    await self._resource_gc_timer_task
                except asyncio.CancelledError:
                    pass
            self._resource_gc_timer_task = None

            if self._state.connection_state != ConnectionState.CLOSED:
                self._state.connection_state = ConnectionState.CLOSED
                self._state.closed_at = get_timestamp()
                close_event_effect = EmitConnectionEvent(
                    event_type=EventType.CONNECTION_CLOSED,
                    data={"connection_id": self._connection_id, "reason": final_reason, "error_code": final_error_code},
                )
                try:
                    self._owner_notify_callback(close_event_effect)
                except Exception as cb_err:
                    logger.warning("Error in final owner notification callback: %s", cb_err)

    async def _early_event_cleanup_loop(self) -> None:
        """Periodically trigger cleanup for early (buffered) events."""
        if self._pending_event_ttl <= 0:
            logger.debug("Early event cleanup timer disabled (pending_event_ttl=0).")
            return

        loop = asyncio.get_running_loop()
        self._next_early_event_cleanup_at = loop.time() + self._pending_event_ttl

        try:
            while self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                now = loop.time()
                wait_time = self._next_early_event_cleanup_at - now

                if wait_time <= 0:
                    if self._state.early_event_count > 0:
                        await self.put_event(event=InternalCleanupEarlyEvents())
                    self._next_early_event_cleanup_at = now + self._pending_event_ttl
                    await asyncio.sleep(delay=self._pending_event_ttl)
                else:
                    await asyncio.sleep(delay=wait_time)

        except asyncio.CancelledError:
            logger.debug("Engine early event cleanup task cancelled.")
        except Exception as e:
            logger.error("Fatal error in engine early event cleanup loop: %s", e, exc_info=True)
            if self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                fut = loop.create_future()
                cleanup_error = ConnectionError(
                    message=f"Early event cleanup loop failed: {e}", error_code=ErrorCodes.INTERNAL_ERROR
                )
                await self.put_event(
                    event=ConnectionClose(error_code=cleanup_error.error_code, reason=cleanup_error.message, future=fut)
                )

    async def _execute_effects(self, *, effects: list[Effect]) -> None:
        """Run the asynchronous side effects returned by the state machine."""
        handler = self._protocol_handler()
        if not handler:
            logger.warning("Protocol handler lost during effect execution, discarding effects.")
            error = ConnectionError("Protocol handler lost")
            for effect in effects:
                match effect:
                    case FailUserFuture(future=fut, exception=err):
                        if not fut.done():
                            try:
                                fut.set_exception(err)
                            except asyncio.InvalidStateError:
                                pass
                    case CompleteUserFuture(future=fut):
                        if not fut.done():
                            try:
                                fut.set_exception(error)
                            except asyncio.InvalidStateError:
                                pass
                    case CreateH3Session(create_future=fut) | CreateQuicStream(create_future=fut):
                        if not fut.done():
                            try:
                                fut.set_exception(error)
                            except asyncio.InvalidStateError:
                                pass
                    case _:
                        pass
            return

        for effect in effects:
            try:
                match effect:
                    case CleanupH3Stream(stream_id=sid):
                        if hasattr(self, "_h3_engine"):
                            self._h3_engine.cleanup_stream(stream_id=sid)

                    case CloseQuicConnection(error_code=ec, reason=r):
                        previous_state = self._state.connection_state
                        if previous_state != ConnectionState.CLOSED:
                            handler.close_connection(error_code=ec, reason_phrase=r)
                            current_state_after_close = self._state.connection_state
                            if current_state_after_close == ConnectionState.CLOSED:
                                self._state.closed_at = get_timestamp()
                                close_event_effect = EmitConnectionEvent(
                                    event_type=EventType.CONNECTION_CLOSED,
                                    data={"connection_id": self._connection_id, "reason": r, "error_code": ec},
                                )
                                self._owner_notify_callback(close_event_effect)
                            elif (
                                current_state_after_close == ConnectionState.CLOSING
                                and previous_state != ConnectionState.CLOSING
                            ):
                                self._state.closed_at = get_timestamp()

                    case RescheduleQuicTimer():
                        handler.schedule_timer_now()

                    case ResetQuicStream(stream_id=sid, error_code=ec):
                        handler.reset_stream(stream_id=sid, error_code=ec)

                    case SendQuicData(stream_id=sid, data=d, end_stream=es):
                        handler.send_stream_data(stream_id=sid, data=d, end_stream=es)

                    case SendQuicDatagram(data=d):
                        handler.send_datagram_frame(data=d)

                    case StopQuicStream(stream_id=sid, error_code=ec):
                        handler.stop_stream(stream_id=sid, error_code=ec)

                    case TriggerQuicTimer():
                        handler.handle_timer_now()

                    case CompleteUserFuture(future=fut, value=v):
                        if not fut.done():
                            fut.set_result(v)

                    case FailUserFuture(future=fut, exception=err):
                        if not fut.done():
                            fut.set_exception(err)

                    case EmitConnectionEvent() | EmitSessionEvent() | EmitStreamEvent() as emit_effect:
                        self._owner_notify_callback(emit_effect)

                    case LogH3Frame(category=c, event=e, data=d):
                        handler.log_event(category=c, event=e, data=d)

                    case CreateH3Session(session_id=sid, path=p, headers=h, create_future=fut):
                        session_data = self._state.sessions.get(sid)
                        if not session_data:
                            if not fut.done():
                                fut.set_exception(SessionError(message=f"Session {sid} vanished during creation"))
                            continue

                        if session_data.control_stream_id != -1:
                            logger.warning(
                                "Session %s already has a control stream ID (%d)", sid, session_data.control_stream_id
                            )
                            if not fut.done():
                                fut.set_exception(SessionError(message=f"Session {sid} control stream already exists"))
                            continue

                        try:
                            control_stream_id = handler.get_next_available_stream_id(is_unidirectional=False)
                        except Exception as e:
                            logger.error("Failed to get next available stream ID for session %s: %s", sid, e)
                            if not fut.done():
                                fut.set_exception(
                                    ConnectionError("No available streams to create session control stream")
                                )
                            continue

                        server_name = handler.get_server_name()
                        if not server_name:
                            error = ConnectionError("Cannot create session: missing server name (SNI)")
                            logger.error(error)
                            if not fut.done():
                                fut.set_exception(error)
                            continue

                        final_headers: Headers = {
                            ":method": "CONNECT",
                            ":scheme": "https",
                            ":authority": server_name,
                            ":path": p,
                            ":protocol": "webtransport",
                        }
                        final_headers.update(h)

                        session_data.control_stream_id = control_stream_id
                        self._state.stream_to_session_map[control_stream_id] = sid
                        self._state.pending_create_session_futures[control_stream_id] = fut

                        h3_effects = self._h3_engine.encode_headers(
                            stream_id=control_stream_id, headers=final_headers, end_stream=False
                        )
                        await self._execute_effects(effects=h3_effects)

                    case CreateQuicStream(session_id=sid, is_unidirectional=is_uni, create_future=fut):
                        session_data = self._state.sessions.get(sid)
                        if session_data:
                            try:
                                actual_stream_id = handler.get_next_available_stream_id(is_unidirectional=is_uni)
                                (h3_effects) = self._h3_engine.encode_webtransport_stream_creation(
                                    stream_id=actual_stream_id,
                                    control_stream_id=session_data.control_stream_id,
                                    is_unidirectional=is_uni,
                                )
                            except Exception as e:
                                logger.error(
                                    "Failed during WebTransport stream creation encoding for session %s: %s", sid, e
                                )
                                if not fut.done():
                                    fut.set_exception(ConnectionError("Failed to encode stream creation frame"))
                                continue

                            stream_data = StreamStateData(
                                stream_id=actual_stream_id,
                                session_id=sid,
                                direction=StreamDirection.SEND_ONLY if is_uni else StreamDirection.BIDIRECTIONAL,
                                state=StreamState.OPEN,
                                created_at=get_timestamp(),
                            )
                            self._state.streams[actual_stream_id] = stream_data
                            self._state.stream_to_session_map[actual_stream_id] = sid
                            await self._execute_effects(effects=h3_effects)
                            if not fut.done():
                                fut.set_result(actual_stream_id)

                            self._owner_notify_callback(
                                EmitStreamEvent(
                                    stream_id=actual_stream_id,
                                    event_type=EventType.STREAM_OPENED,
                                    data={
                                        "stream_id": actual_stream_id,
                                        "session_id": sid,
                                        "direction": stream_data.direction,
                                    },
                                )
                            )
                        else:
                            if not fut.done():
                                fut.set_exception(
                                    SessionError(message=f"Session {sid} vanished during stream creation")
                                )

                    case SendH3Capsule(stream_id=sid, capsule_type=ct, capsule_data=cd):
                        encoded_bytes = self._h3_engine.encode_capsule(stream_id=sid, capsule_type=ct, capsule_data=cd)
                        await self._execute_effects(
                            effects=[SendQuicData(stream_id=sid, data=encoded_bytes, end_stream=False)]
                        )

                    case SendH3Datagram(stream_id=sid, data=d):
                        encoded_bytes = self._h3_engine.encode_datagram(stream_id=sid, data=d)
                        await self._execute_effects(effects=[SendQuicDatagram(data=encoded_bytes)])

                    case SendH3Goaway():
                        h3_control_stream_id = self._h3_engine._local_control_stream_id
                        if h3_control_stream_id is None:
                            logger.critical("Cannot send GOAWAY: H3 local control stream ID is None.")
                        else:
                            goaway_bytes = self._h3_engine.encode_goaway_frame()
                            await self._execute_effects(
                                effects=[
                                    SendQuicData(stream_id=h3_control_stream_id, data=goaway_bytes, end_stream=False)
                                ]
                            )

                    case SendH3Headers(stream_id=sid, status=st, end_stream=es):
                        h3_effects = self._h3_engine.encode_headers(
                            stream_id=sid, headers={":status": str(st)}, end_stream=es
                        )
                        await self._execute_effects(effects=h3_effects)

                    case _:
                        logger.warning("Unhandled effect type in execute_effects: %s", type(effect))

            except Exception as e:
                logger.error("Error executing effect %s: %s", type(effect).__name__, e, exc_info=True)
                if self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                    close_effects_on_error: list[Effect] = [
                        CloseQuicConnection(error_code=ErrorCodes.INTERNAL_ERROR, reason="Effect execution error")
                    ]
                    try:
                        await self._execute_effects(effects=close_effects_on_error)
                    except Exception as close_e:
                        logger.critical("Failed even to close connection after effect error: %s", close_e)
                        if self._state.connection_state != ConnectionState.CLOSED:
                            self._state.connection_state = ConnectionState.CLOSED
                            self._state.closed_at = get_timestamp()
                            final_close_event = EmitConnectionEvent(
                                event_type=EventType.CONNECTION_CLOSED,
                                data={
                                    "connection_id": self._connection_id,
                                    "reason": "Effect execution error, close failed",
                                    "error_code": ErrorCodes.INTERNAL_ERROR,
                                },
                            )
                            try:
                                self._owner_notify_callback(final_close_event)
                            except Exception:
                                pass

    def _handle_event(self, *, event: ProtocolEvent) -> list[Effect]:
        """Process a single event and return resulting effects."""
        all_effects: list[Effect] = []
        events_to_process: deque[ProtocolEvent] = deque([event])

        while events_to_process:
            current_event = events_to_process.popleft()
            new_effects: list[Effect] = []
            re_queue_pending_actions = False

            match current_event:
                case InternalCleanupEarlyEvents():
                    if self._state.early_event_count == 0:
                        break

                    now = get_timestamp()
                    expired_control_stream_ids = [
                        control_stream_id
                        for control_stream_id, events in self._state.early_event_buffer.items()
                        if events and (now - events[0][0]) > self._pending_event_ttl
                    ]

                    for control_stream_id in expired_control_stream_ids:
                        events_to_discard = self._state.early_event_buffer.pop(control_stream_id, [])
                        if not events_to_discard:
                            continue

                        self._state.early_event_count -= len(events_to_discard)
                        logger.warning(
                            "Discarding %d expired early events for unknown session (control stream %d)",
                            len(events_to_discard),
                            control_stream_id,
                        )
                        for _ts, ev in events_to_discard:
                            if isinstance(ev, WebTransportStreamDataReceived):
                                new_effects.append(
                                    ResetQuicStream(
                                        stream_id=ev.stream_id, error_code=ErrorCodes.WT_BUFFERED_STREAM_REJECTED
                                    )
                                )

                case InternalCleanupResources() as icr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_cleanup_resources(event=icr_ev, state=self._state)
                    )

                case TransportConnectionTerminated() as tct_ev:
                    new_effects.extend(
                        self._connection_processor.handle_connection_terminated(event=tct_ev, state=self._state)
                    )

                case TransportDatagramFrameReceived() | TransportStreamDataReceived() as tev:
                    was_settings_received = self._h3_engine._settings_received

                    h3_events, h3_effects = self._h3_engine.handle_transport_event(event=tev, state=self._state)
                    new_effects.extend(h3_effects)
                    events_to_process.extendleft(reversed(h3_events))

                    if self._is_client and not was_settings_received and self._h3_engine._settings_received:
                        logger.debug("Client received peer H3 SETTINGS.")
                        self._state.peer_settings_received = True
                        (readiness_effects, is_ready) = self._check_client_connection_ready()
                        new_effects.extend(readiness_effects)
                        if is_ready:
                            re_queue_pending_actions = True

                case TransportHandshakeCompleted():
                    if self._state.connection_state == ConnectionState.CONNECTING:
                        logger.debug("TransportHandshakeCompleted received. Initializing H3.")
                        handler = self._protocol_handler()
                        if not handler:
                            logger.error("Protocol handler lost during H3 initialization.")
                            new_effects.append(
                                CloseQuicConnection(
                                    error_code=ErrorCodes.INTERNAL_ERROR, reason="Protocol handler lost"
                                )
                            )
                            break

                        try:
                            control_stream_id = handler.get_next_available_stream_id(is_unidirectional=True)
                            encoder_stream_id = handler.get_next_available_stream_id(is_unidirectional=True)
                            decoder_stream_id = handler.get_next_available_stream_id(is_unidirectional=True)

                            self._h3_engine.set_local_stream_ids(
                                control_stream_id=control_stream_id,
                                encoder_stream_id=encoder_stream_id,
                                decoder_stream_id=decoder_stream_id,
                            )

                            settings_bytes = self._h3_engine.initialize_connection()

                            new_effects = [
                                SendQuicData(
                                    stream_id=control_stream_id,
                                    data=encode_uint_var(constants.H3_STREAM_TYPE_CONTROL) + settings_bytes,
                                    end_stream=False,
                                ),
                                SendQuicData(
                                    stream_id=encoder_stream_id,
                                    data=encode_uint_var(constants.H3_STREAM_TYPE_QPACK_ENCODER),
                                    end_stream=False,
                                ),
                                SendQuicData(
                                    stream_id=decoder_stream_id,
                                    data=encode_uint_var(constants.H3_STREAM_TYPE_QPACK_DECODER),
                                    end_stream=False,
                                ),
                                LogH3Frame(
                                    category="http",
                                    event="stream_type_set",
                                    data={"new": "control", "stream_id": control_stream_id},
                                ),
                                LogH3Frame(
                                    category="http",
                                    event="stream_type_set",
                                    data={"new": "qpack_encoder", "stream_id": encoder_stream_id},
                                ),
                                LogH3Frame(
                                    category="http",
                                    event="stream_type_set",
                                    data={"new": "qpack_decoder", "stream_id": decoder_stream_id},
                                ),
                            ]
                            self._state.handshake_complete = True

                            if self._is_client:
                                (readiness_effects, is_ready) = self._check_client_connection_ready()
                                new_effects.extend(readiness_effects)
                                if is_ready:
                                    re_queue_pending_actions = True
                            else:
                                self._state.connection_state = ConnectionState.CONNECTED
                                self._state.connected_at = get_timestamp()
                                new_effects.append(
                                    EmitConnectionEvent(
                                        event_type=EventType.CONNECTION_ESTABLISHED,
                                        data={"connection_id": self._connection_id},
                                    )
                                )
                                logger.debug("Server connection established and H3 initialized.")

                        except Exception as h3_init_error:
                            logger.error("H3 initialization failed after handshake: %s", h3_init_error, exc_info=True)
                            new_effects.append(
                                CloseQuicConnection(
                                    error_code=ErrorCodes.INTERNAL_ERROR,
                                    reason=f"H3 initialization failed: {h3_init_error}",
                                )
                            )
                    else:
                        logger.warning(
                            "Received TransportHandshakeCompleted in unexpected state: %s", self._state.connection_state
                        )

                case TransportQuicParametersReceived() as tqpr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_transport_parameters_received(
                            event=tqpr_ev, state=self._state
                        )
                    )

                case TransportQuicTimerFired():
                    new_effects.extend([TriggerQuicTimer(), RescheduleQuicTimer()])

                case TransportStreamReset() as tsr_ev:
                    new_effects.extend(
                        self._stream_processor.handle_transport_stream_reset(event=tsr_ev, state=self._state)
                    )

                case CapsuleReceived() as cr_ev:
                    new_effects.extend(self._session_processor.handle_capsule_received(event=cr_ev, state=self._state))

                case ConnectStreamClosed() as csc_ev:
                    new_effects.extend(
                        self._session_processor.handle_connect_stream_closed(event=csc_ev, state=self._state)
                    )

                case DatagramReceived() as dr_ev:
                    new_effects.extend(self._session_processor.handle_datagram_received(event=dr_ev, state=self._state))

                case GoawayReceived() as gr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_goaway_received(event=gr_ev, state=self._state)
                    )

                case HeadersReceived() as hr_ev:
                    new_effects.extend(
                        self._connection_processor.handle_headers_received(event=hr_ev, state=self._state)
                    )

                    session_id_to_check: SessionId | None = None
                    for effect in new_effects:
                        if isinstance(effect, EmitSessionEvent) and effect.event_type in (
                            EventType.SESSION_REQUEST,
                            EventType.SESSION_READY,
                        ):
                            session_id_to_check = effect.session_id
                            break

                    if session_id_to_check:
                        session_data = self._state.sessions.get(session_id_to_check)
                        if session_data:
                            control_stream_id = session_data.control_stream_id
                            buffered_events = self._state.early_event_buffer.pop(control_stream_id, None)
                            if buffered_events:
                                logger.debug(
                                    "Session %s (ControlStream %d) is now active, re-queueing %d buffered events.",
                                    session_id_to_check,
                                    control_stream_id,
                                    len(buffered_events),
                                )
                                self._state.early_event_count -= len(buffered_events)
                                events_to_process.extendleft(reversed([evt for _ts, evt in buffered_events]))

                case SettingsReceived() as sr_ev:
                    logger.debug("Processing H3 SETTINGS frame.")
                    self._state.peer_initial_max_data = sr_ev.settings.get(constants.SETTINGS_WT_INITIAL_MAX_DATA, 0)
                    self._state.peer_initial_max_streams_bidi = sr_ev.settings.get(
                        constants.SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI, 0
                    )
                    self._state.peer_initial_max_streams_uni = sr_ev.settings.get(
                        constants.SETTINGS_WT_INITIAL_MAX_STREAMS_UNI, 0
                    )

                case WebTransportStreamDataReceived() as wtsdr_ev:
                    new_effects.extend(
                        self._stream_processor.handle_webtransport_stream_data(event=wtsdr_ev, state=self._state)
                    )

                case UserAcceptSession() as uas_ev:
                    new_effects.extend(self._session_processor.handle_accept_session(event=uas_ev, state=self._state))

                case UserCloseSession() as ucs_ev:
                    new_effects.extend(self._session_processor.handle_close_session(event=ucs_ev, state=self._state))

                case UserConnectionGracefulClose() as ugcc_ev:
                    new_effects.extend(
                        self._connection_processor.handle_graceful_close(event=ugcc_ev, state=self._state)
                    )

                case UserCreateSession() as ucs_ev:
                    if self._is_client and self._state.connection_state != ConnectionState.CONNECTED:
                        logger.debug("Client not fully connected, buffering UserCreateSession.")
                        self._pending_user_actions.append(ucs_ev)
                    else:
                        new_effects.extend(
                            self._connection_processor.handle_create_session(event=ucs_ev, state=self._state)
                        )

                case UserCreateStream() as ucs_ev:
                    if self._is_client and self._state.connection_state != ConnectionState.CONNECTED:
                        logger.debug("Client not fully connected, buffering UserCreateStream.")
                        self._pending_user_actions.append(ucs_ev)
                    else:
                        new_effects.extend(
                            self._session_processor.handle_create_stream(event=ucs_ev, state=self._state)
                        )

                case UserGetConnectionDiagnostics() as ugcd_ev:
                    new_effects.extend(
                        self._connection_processor.handle_get_connection_diagnostics(event=ugcd_ev, state=self._state)
                    )

                case UserGetSessionDiagnostics() as ugsd_ev:
                    new_effects.extend(
                        self._session_processor.handle_get_session_diagnostics(event=ugsd_ev, state=self._state)
                    )

                case UserGetStreamDiagnostics() as ugstd_ev:
                    new_effects.extend(
                        self._stream_processor.handle_get_stream_diagnostics(event=ugstd_ev, state=self._state)
                    )

                case UserGrantDataCredit() as ugdc_ev:
                    new_effects.extend(
                        self._session_processor.handle_grant_data_credit(event=ugdc_ev, state=self._state)
                    )

                case UserGrantStreamsCredit() as ugsc_ev:
                    new_effects.extend(
                        self._session_processor.handle_grant_streams_credit(event=ugsc_ev, state=self._state)
                    )

                case UserRejectSession() as urs_ev:
                    new_effects.extend(self._session_processor.handle_reject_session(event=urs_ev, state=self._state))

                case UserResetStream() as urs_ev:
                    new_effects.extend(self._stream_processor.handle_reset_stream(event=urs_ev, state=self._state))

                case UserSendDatagram() as usd_ev:
                    new_effects.extend(self._session_processor.handle_send_datagram(event=usd_ev, state=self._state))

                case UserSendStreamData() as ussd_ev:
                    new_effects.extend(self._stream_processor.handle_send_stream_data(event=ussd_ev, state=self._state))

                case UserStopStream() as uss_ev:
                    new_effects.extend(self._stream_processor.handle_stop_stream(event=uss_ev, state=self._state))

                case UserStreamRead() as usr_ev:
                    new_effects.extend(self._stream_processor.handle_stream_read(event=usr_ev, state=self._state))

                case ConnectionClose() as cc_ev:
                    new_effects.extend(
                        self._connection_processor.handle_connection_close(event=cc_ev, state=self._state)
                    )

                case _:
                    logger.warning("Unhandled event type in engine's handle_event: %s", type(current_event))

            all_effects.extend(new_effects)

            if re_queue_pending_actions:
                if self._pending_user_actions:
                    logger.debug(
                        "Connection is ready, re-queueing %d pending user actions.", len(self._pending_user_actions)
                    )
                    events_to_process.extendleft(reversed(self._pending_user_actions))
                    self._pending_user_actions.clear()

        if not any(isinstance(effect, RescheduleQuicTimer) for effect in all_effects):
            all_effects.append(RescheduleQuicTimer())

        return all_effects

    async def _resource_gc_timer_loop(self) -> None:
        """Periodically trigger garbage collection for closed resources."""
        if self._resource_cleanup_interval <= 0:
            logger.debug("Resource GC timer disabled (interval=0).")
            return

        loop = asyncio.get_running_loop()
        try:
            while self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                await asyncio.sleep(delay=self._resource_cleanup_interval)

                if self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                    await self.put_event(event=InternalCleanupResources())

        except asyncio.CancelledError:
            logger.debug("Engine resource GC timer task cancelled.")
        except Exception as e:
            logger.error("Fatal error in engine resource GC timer loop: %s", e, exc_info=True)
            if self._state.connection_state not in (ConnectionState.CLOSING, ConnectionState.CLOSED):
                fut = loop.create_future()
                timer_error = ConnectionError(
                    message=f"Resource GC loop failed: {e}", error_code=ErrorCodes.INTERNAL_ERROR
                )
                await self.put_event(
                    event=ConnectionClose(error_code=timer_error.error_code, reason=timer_error.message, future=fut)
                )



"""Shared, general-purpose utilities."""

from __future__ import annotations

import logging
import secrets
import time
from datetime import datetime, timedelta, timezone
from pathlib import Path
from types import TracebackType
from typing import Self

from aioquic.quic.configuration import QuicConfiguration
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from pywebtransport.types import Buffer

__all__: list[str] = [
    "Timer",
    "create_quic_configuration",
    "ensure_bytes",
    "format_duration",
    "generate_self_signed_cert",
    "generate_session_id",
    "get_logger",
    "get_timestamp",
]


class Timer:
    """A simple context manager for performance measurement."""

    def __init__(self, *, name: str = "timer") -> None:
        """Initialize the timer."""
        self.name = name
        self.start_time: float | None = None
        self.end_time: float | None = None

    @property
    def elapsed(self) -> float:
        """Get the elapsed time in seconds."""
        if self.start_time is None:
            return 0.0

        end = self.end_time or time.time()
        return end - self.start_time

    def __enter__(self) -> Self:
        """Start the timer upon entering the context."""
        self.start()
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Stop the timer and log the duration upon exiting the context."""
        elapsed = self.stop()
        logger = get_logger(name="timer")
        logger.debug("%s took %s", self.name, format_duration(seconds=elapsed))

    def start(self) -> None:
        """Start the timer."""
        self.start_time = time.time()
        self.end_time = None

    def stop(self) -> float:
        """Stop the timer and return the elapsed time."""
        if self.start_time is None:
            raise RuntimeError("Timer not started")

        self.end_time = time.time()
        return self.elapsed


def create_quic_configuration(
    *,
    alpn_protocols: list[str],
    congestion_control_algorithm: str,
    idle_timeout: float,
    is_client: bool,
    max_datagram_size: int,
) -> QuicConfiguration:
    """Create a QUIC configuration from specific, required parameters."""
    return QuicConfiguration(
        alpn_protocols=alpn_protocols,
        congestion_control_algorithm=congestion_control_algorithm,
        idle_timeout=idle_timeout,
        is_client=is_client,
        max_datagram_frame_size=max_datagram_size,
    )


def ensure_bytes(*, data: Buffer | str, encoding: str = "utf-8") -> bytes:
    """Ensure that the given data is in bytes format."""
    match data:
        case str():
            return data.encode(encoding)
        case bytes():
            return data
        case bytearray() | memoryview():
            return bytes(data)
        case _:
            raise TypeError(f"Expected str, bytes, bytearray, or memoryview, got {type(data).__name__}")


def format_duration(*, seconds: float) -> str:
    """Format a duration in seconds into a human-readable string."""
    if seconds < 1:
        return f"{seconds * 1000:.1f}ms"
    if seconds < 60:
        return f"{seconds:.1f}s"
    if seconds < 3600:
        minutes = int(seconds // 60)
        secs = seconds % 60
        return f"{minutes}m{secs:.1f}s"

    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    secs = seconds % 60
    return f"{hours}h{minutes}m{secs:.1f}s"


def generate_self_signed_cert(
    *, hostname: str, output_dir: str = ".", days_valid: int = 365, key_size: int = 2048
) -> tuple[str, str]:
    """Generate a self-signed certificate and key for testing purposes."""
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "CA"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "pywebtransport"),
            x509.NameAttribute(NameOID.COMMON_NAME, hostname),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=days_valid))
        .add_extension(x509.SubjectAlternativeName([x509.DNSName(hostname)]), critical=False)
        .sign(private_key, hashes.SHA256())
    )

    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    cert_file = output_path / f"{hostname}.crt"
    key_file = output_path / f"{hostname}.key"

    with open(cert_file, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))
    with open(key_file, "wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    return (str(cert_file), str(key_file))


def generate_session_id() -> str:
    """Generate a unique, URL-safe session ID."""
    return secrets.token_urlsafe(16)


def get_logger(*, name: str) -> logging.Logger:
    """Get a logger instance with a specific name."""
    return logging.getLogger(name)


def get_timestamp() -> float:
    """Get the current Unix timestamp."""
    return time.time()



"""Core server implementation for accepting WebTransport connections."""

from __future__ import annotations

import asyncio
from asyncio import BaseTransport, DatagramTransport
from collections import Counter
from dataclasses import asdict, dataclass
from pathlib import Path
from types import TracebackType
from typing import Any, Self, cast

from aioquic.asyncio.server import QuicServer

from pywebtransport.config import ServerConfig
from pywebtransport.connection.connection import WebTransportConnection
from pywebtransport.events import Event, EventEmitter
from pywebtransport.exceptions import ServerError
from pywebtransport.manager.connection import ConnectionManager
from pywebtransport.manager.session import SessionManager
from pywebtransport.server._adapter import WebTransportServerProtocol, create_server
from pywebtransport.types import Address, ConnectionState, EventType, SessionState
from pywebtransport.utils import get_logger, get_timestamp

__all__ = ["ServerDiagnostics", "ServerStats", "WebTransportServer"]

logger = get_logger(name=__name__)


@dataclass(frozen=True, kw_only=True)
class ServerDiagnostics:
    """A structured, immutable snapshot of a server's health."""

    stats: ServerStats
    connection_states: dict[ConnectionState, int]
    session_states: dict[SessionState, int]
    is_serving: bool
    certfile_path: str
    keyfile_path: str
    max_connections: int

    @property
    def issues(self) -> list[str]:
        """Get a list of potential issues based on the current diagnostics."""
        issues: list[str] = []
        stats_dict = self.stats.to_dict()

        if not self.is_serving:
            issues.append("Server is not currently serving.")

        total_attempts = stats_dict.get("total_connections_attempted", 0)
        success_rate = stats_dict.get("success_rate", 1.0)
        connections_rejected = stats_dict.get("connections_rejected", 0)

        if total_attempts > 20 and success_rate < 0.9:
            issues.append(f"High connection rejection rate: {connections_rejected}/{total_attempts}")

        active_connections = self.connection_states.get(ConnectionState.CONNECTED, 0)
        if self.max_connections > 0 and (active_connections / max(1, self.max_connections)) > 0.9:
            issues.append(f"High connection usage: {active_connections / self.max_connections:.1%}")

        try:
            cert_exists = Path(self.certfile_path).exists() if self.certfile_path else False
            key_exists = Path(self.keyfile_path).exists() if self.keyfile_path else False
            if not cert_exists:
                issues.append(f"Certificate file not found: {self.certfile_path}")
            if not key_exists:
                issues.append(f"Key file not found: {self.keyfile_path}")
        except Exception as e:
            issues.append(f"Certificate configuration check failed: {e}")

        return issues


@dataclass(kw_only=True)
class ServerStats:
    """Represent statistics for the server."""

    start_time: float | None = None
    connections_accepted: int = 0
    connections_rejected: int = 0
    connection_errors: int = 0
    protocol_errors: int = 0

    @property
    def total_connections_attempted(self) -> int:
        """Get the total number of connections attempted."""
        return self.connections_accepted + self.connections_rejected

    @property
    def success_rate(self) -> float:
        """Get the connection success rate."""
        total = self.total_connections_attempted
        if total == 0:
            return 1.0
        return self.connections_accepted / total

    def to_dict(self) -> dict[str, Any]:
        """Convert statistics to a dictionary."""
        data = asdict(obj=self)
        data["total_connections_attempted"] = self.total_connections_attempted
        data["success_rate"] = self.success_rate
        data["uptime"] = (get_timestamp() - self.start_time) if self.start_time else 0.0
        return data


class WebTransportServer(EventEmitter):
    """Manage the lifecycle and connections for the WebTransport server."""

    def __init__(self, *, config: ServerConfig | None = None) -> None:
        """Initialize the WebTransport server."""
        super().__init__()
        self._config = config or ServerConfig()
        self._config.validate()
        self._serving, self._closing = False, False
        self._server: QuicServer | None = None
        self._connection_manager = ConnectionManager(max_connections=self._config.max_connections)
        self._session_manager = SessionManager(max_sessions=self._config.max_sessions)
        self._background_tasks: list[asyncio.Task[Any]] = []
        self._stats = ServerStats()
        self._shutdown_event: asyncio.Event | None = None
        logger.info("WebTransport server initialized.")

    @property
    def config(self) -> ServerConfig:
        """Get the server's configuration object."""
        return self._config

    @property
    def connection_manager(self) -> ConnectionManager:
        """Get the server's connection manager instance."""
        return self._connection_manager

    @property
    def is_serving(self) -> bool:
        """Check if the server is currently serving."""
        return self._serving

    @property
    def local_address(self) -> Address | None:
        """Get the local address the server is bound to."""
        if self._server and hasattr(self._server, "_transport") and self._server._transport:
            try:
                return cast(Address | None, self._server._transport.get_extra_info("sockname"))
            except OSError:
                return None
        return None

    @property
    def session_manager(self) -> SessionManager:
        """Get the server's session manager instance."""
        return self._session_manager

    async def __aenter__(self) -> Self:
        """Enter the async context for the server."""
        await self._connection_manager.__aenter__()
        await self._session_manager.__aenter__()
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and close the server."""
        await self.close()

    async def close(self) -> None:
        """Gracefully shut down the server and its resources."""
        if not self._serving:
            return

        logger.info("Closing WebTransport server...")
        self._serving = False
        self._closing = True

        if self._shutdown_event:
            self._shutdown_event.set()

        for task in self._background_tasks:
            if not task.done():
                task.cancel()
        if self._background_tasks:
            await asyncio.gather(*self._background_tasks, return_exceptions=True)

        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(coro=self._connection_manager.shutdown())
                tg.create_task(coro=self._session_manager.shutdown())
        except* Exception as eg:
            logger.error("Errors occurred during manager shutdown: %s", eg.exceptions, exc_info=eg)

        if self._server:
            self._server.close()

        self._closing = False
        logger.info("WebTransport server closed.")

    async def diagnostics(self) -> ServerDiagnostics:
        """Get a snapshot of the server's diagnostics and statistics."""
        async with asyncio.TaskGroup() as tg:
            conn_task = tg.create_task(coro=self._connection_manager.get_all_resources())
            sess_task = tg.create_task(coro=self._session_manager.get_all_resources())

        connections = conn_task.result()
        sessions = sess_task.result()
        connection_states = Counter(conn.state for conn in connections)
        session_states = Counter(sess.state for sess in sessions)

        return ServerDiagnostics(
            stats=self._stats,
            connection_states=dict(connection_states),
            session_states=dict(session_states),
            is_serving=self.is_serving,
            certfile_path=self.config.certfile or "",
            keyfile_path=self.config.keyfile or "",
            max_connections=self.config.max_connections,
        )

    async def listen(self, *, host: str | None = None, port: int | None = None) -> None:
        """Start the server and begin listening for connections."""
        if self._serving:
            raise ServerError(message="Server is already serving")

        bind_host, bind_port = host or self._config.bind_host, port or self._config.bind_port
        logger.info("Starting WebTransport server on %s:%s", bind_host, bind_port)

        self._shutdown_event = asyncio.Event()

        try:
            self._server = await create_server(
                host=bind_host, port=bind_port, config=self._config, connection_creator=self._create_connection_callback
            )
            self._serving = True
            self._stats.start_time = get_timestamp()
            logger.info("WebTransport server listening on %s", self.local_address)
        except FileNotFoundError as e:
            logger.critical("Certificate/Key file error: %s", e)
            raise ServerError(message=f"Certificate/Key file error: {e}") from e
        except Exception as e:
            logger.critical("Failed to start server: %s", e, exc_info=True)
            raise ServerError(message=f"Failed to start server: {e}") from e

    async def serve_forever(self) -> None:
        """Run the server indefinitely until interrupted."""
        if not self._serving or not self._server:
            raise ServerError(message="Server is not listening")

        logger.info("Server is running. Press Ctrl+C to stop.")
        try:
            if self._shutdown_event:
                await self._shutdown_event.wait()
        except asyncio.CancelledError:
            logger.info("serve_forever cancelled.")
        except Exception as e:
            logger.error("Error during serve_forever wait: %s", e)
        finally:
            logger.info("serve_forever loop finished.")

    def _create_connection_callback(
        self, protocol: WebTransportServerProtocol, transport: BaseTransport
    ) -> WebTransportConnection | None:
        """Create a new WebTransportConnection from the protocol."""
        logger.debug("Creating WebTransportConnection via callback.")

        if not hasattr(transport, "sendto"):
            logger.error("Received transport without 'sendto' method: %s", type(transport).__name__)
            if hasattr(transport, "close"):
                if not hasattr(transport, "is_closing") or not transport.is_closing():
                    transport.close()
            return None

        try:
            connection = WebTransportConnection(
                config=self._config, protocol=protocol, transport=cast(DatagramTransport, transport), is_client=False
            )
            asyncio.create_task(coro=self._initialize_and_register_connection(connection=connection))
            return connection
        except Exception as e:
            logger.error("Error creating WebTransportConnection in callback: %s", e, exc_info=True)
            if not transport.is_closing():
                transport.close()
            return None

    async def _initialize_and_register_connection(self, connection: WebTransportConnection) -> None:
        """Initialize connection engine and register with manager."""
        try:

            async def forward_session_request(event: Event) -> None:
                event_data = event.data.copy() if isinstance(event.data, dict) else {}
                event_data["connection"] = connection
                await self.emit(event_type=EventType.SESSION_REQUEST, data=event_data)

            connection.events.on(event_type=EventType.SESSION_REQUEST, handler=forward_session_request)

            await connection.initialize()

            await self._connection_manager.add_connection(connection=connection)
            self._stats.connections_accepted += 1
            logger.info("New connection registered: %s", connection.connection_id)
        except Exception as e:
            self._stats.connections_rejected += 1
            self._stats.connection_errors += 1
            logger.error("Failed to initialize/register new connection: %s", e, exc_info=True)
            if not connection.is_closed:
                await connection.close()

    def __str__(self) -> str:
        """Format a concise summary of server information for logging."""
        status = "serving" if self.is_serving else "stopped"
        address_info = self.local_address
        address_str = f"{address_info[0]}:{address_info[1]}" if address_info else "unknown"
        conn_count = len(self._connection_manager)
        sess_count = len(self._session_manager)
        return (
            f"WebTransportServer(status={status}, "
            f"address={address_str}, "
            f"connections={conn_count}, "
            f"sessions={sess_count})"
        )



"""Robust pool for managing WebTransportSession objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pywebtransport.pool._base import _AsyncObjectPool
from pywebtransport.session.session import WebTransportSession

if TYPE_CHECKING:
    from pywebtransport.client.client import WebTransportClient
    from pywebtransport.types import URL


__all__: list[str] = ["SessionPool"]


class SessionPool(_AsyncObjectPool[WebTransportSession]):
    """A robust pool for reusing and managing concurrent sessions for a client and URL."""

    def __init__(self, *, client: WebTransportClient, url: URL, max_size: int) -> None:
        """Initialize the session pool."""

        async def factory() -> WebTransportSession:
            return await client.connect(url=url)

        super().__init__(max_size=max_size, factory=factory)

    async def _dispose(self, session: WebTransportSession) -> None:
        """Close and dispose of a single session object."""
        if not session.is_closed:
            await session.close()



"""Generic, reusable asynchronous object pool."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import AsyncContextManager, Generic, Self, TypeVar

__all__: list[str] = []

T = TypeVar("T")


class _AsyncObjectPool(ABC, Generic[T]):
    """A generic, robust asynchronous object pool implementation."""

    def __init__(self, *, max_size: int, factory: Callable[[], Awaitable[T]]) -> None:
        """Initialize the asynchronous object pool."""
        if max_size <= 0:
            raise ValueError("Pool max_size must be a positive integer.")

        self._factory = factory
        self._max_size = max_size
        self._lock: asyncio.Lock | None = None
        self._semaphore: asyncio.Semaphore | None = None
        self._pool: deque[T] = deque()
        self._active_count = 0
        self._closed = False

    async def __aenter__(self) -> Self:
        """Enter the async context and initialize async resources."""
        if self._lock is None:
            self._lock = asyncio.Lock()
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(value=self._max_size)
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and close the pool."""
        await self.close()

    async def close(self) -> None:
        """Close the pool and dispose of all currently available objects."""
        if self._closed:
            return

        self._closed = True

        objects_to_dispose: list[T] = []
        if self._lock:
            async with self._lock:
                objects_to_dispose = list(self._pool)
                self._pool.clear()
        else:
            objects_to_dispose = list(self._pool)
            self._pool.clear()

        if objects_to_dispose:
            try:
                async with asyncio.TaskGroup() as tg:
                    for obj in objects_to_dispose:
                        tg.create_task(coro=self._dispose(obj))
            except* Exception as eg:
                raise RuntimeError(f"Errors occurred during pool cleanup: {eg.exceptions}") from eg

    async def acquire(self) -> T:
        """Acquire an object from the pool, creating a new one if necessary."""
        if self._closed:
            raise RuntimeError("Cannot acquire from a closed pool.")
        if self._lock is None or self._semaphore is None:
            raise RuntimeError(
                "_AsyncObjectPool has not been activated. It must be used as an "
                "asynchronous context manager (`async with ...`)."
            )

        await self._semaphore.acquire()

        async with self._lock:
            if self._pool:
                return self._pool.popleft()

        try:
            new_obj = await self._factory()
            async with self._lock:
                self._active_count += 1
            return new_obj
        except Exception:
            self._semaphore.release()
            raise

    def get(self) -> AsyncContextManager[T]:
        """Return an async context manager for acquiring and automatically releasing an object."""
        return _PooledObject(pool=self)

    async def release(self, obj: T) -> None:
        """Release an object, returning it to the pool."""
        if self._lock is None or self._semaphore is None:
            if self._closed:
                await self._dispose(obj)
                return
            raise RuntimeError(
                "_AsyncObjectPool has not been activated. It must be used as an "
                "asynchronous context manager (`async with ...`)."
            )

        if self._closed:
            await self._dispose(obj)
            return

        async with self._lock:
            self._pool.append(obj)

        self._semaphore.release()

    @abstractmethod
    async def _dispose(self, obj: T) -> None:
        """Dispose of a pooled object. Subclasses must implement this."""
        raise NotImplementedError


class _PooledObject(AsyncContextManager[T]):
    """An async context manager for safely acquiring and releasing a pooled object."""

    def __init__(self, pool: _AsyncObjectPool[T]) -> None:
        """Initialize the pooled object context manager."""
        self._pool = pool
        self._obj: T | None = None

    async def __aenter__(self) -> T:
        """Acquire the object from the pool."""
        self._obj = await self._pool.acquire()
        return self._obj

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Release the object back to the pool."""
        if self._obj is not None:
            await self._pool.release(self._obj)



"""Generic base class for monitoring components."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Awaitable
from types import TracebackType
from typing import Any, Generic, Self, TypeVar

from pywebtransport.utils import get_logger

__all__: list[str] = []

logger = get_logger(name=__name__)

T = TypeVar("T")


class _BaseMonitor(ABC, Generic[T]):
    """Implement asynchronous monitoring component base class."""

    def __init__(
        self, target: T, *, monitoring_interval: float = 30.0, metrics_maxlen: int = 120, alerts_maxlen: int = 100
    ) -> None:
        """Initialize the base monitor."""
        self._target = target
        self._interval = monitoring_interval
        self._monitor_task: asyncio.Task[None] | None = None
        self._metrics_history: deque[dict[str, Any]] = deque(maxlen=metrics_maxlen)
        self._alerts: deque[dict[str, Any]] = deque(maxlen=alerts_maxlen)

    @property
    def is_monitoring(self) -> bool:
        """Check if the monitoring task is currently active."""
        return self._monitor_task is not None and not self._monitor_task.done()

    async def __aenter__(self) -> Self:
        """Enter the async context and start the monitoring task."""
        if self.is_monitoring:
            return self

        try:
            self._monitor_task = asyncio.create_task(coro=self._monitor_loop())
            logger.info("%s monitoring started.", self.__class__.__name__)
        except RuntimeError:
            logger.error("Failed to start %s: No running event loop.", self.__class__.__name__, exc_info=True)
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and stop the monitoring task."""
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
        logger.info("%s monitoring stopped.", self.__class__.__name__)

    def clear_history(self) -> None:
        """Clear all collected metrics and alerts history."""
        self._metrics_history.clear()
        self._alerts.clear()
        logger.info("Metrics and alerts history cleared for %s.", self.__class__.__name__)

    def get_alerts(self, *, limit: int = 25) -> list[dict[str, Any]]:
        """Get a list of recently generated alerts."""
        return list(self._alerts)[-limit:]

    def get_metrics_history(self, *, limit: int = 100) -> list[dict[str, Any]]:
        """Get a list of recent metrics history."""
        return list(self._metrics_history)[-limit:]

    @abstractmethod
    def _check_for_alerts(self) -> None | Awaitable[None]:
        """Analyze metrics and generate alerts if thresholds are breached."""
        raise NotImplementedError

    @abstractmethod
    def _collect_metrics(self) -> None | Awaitable[None]:
        """Collect a snapshot of the target's current statistics."""
        raise NotImplementedError

    async def _monitor_loop(self) -> None:
        """Run the main loop for periodically collecting metrics and checking health."""
        try:
            await asyncio.sleep(delay=0)
            while True:
                collection_result = self._collect_metrics()
                if asyncio.iscoroutine(collection_result):
                    await collection_result

                alert_check_result = self._check_for_alerts()
                if asyncio.iscoroutine(alert_check_result):
                    await alert_check_result

                await asyncio.sleep(delay=self._interval)
        except asyncio.CancelledError:
            logger.info("%s loop has been cancelled.", self.__class__.__name__)
        except Exception as e:
            logger.error("%s loop encountered a critical error: %s", self.__class__.__name__, e, exc_info=True)



"""Publish-Subscribe messaging pattern manager."""

from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from types import TracebackType
from typing import Self

from pywebtransport.constants import ErrorCodes
from pywebtransport.exceptions import ConnectionError, StreamError
from pywebtransport.pubsub.exceptions import NotSubscribedError, PubSubError, SubscriptionFailedError
from pywebtransport.stream.stream import WebTransportStream
from pywebtransport.types import Data
from pywebtransport.utils import ensure_bytes, get_logger, get_timestamp

__all__: list[str] = ["PubSubManager", "PubSubStats", "Subscription"]

logger = get_logger(name=__name__)


@dataclass(kw_only=True)
class PubSubStats:
    """Represents statistics for the Pub/Sub manager."""

    created_at: float = field(default_factory=get_timestamp)
    topics_subscribed: int = 0
    messages_published: int = 0
    messages_received: int = 0
    subscription_errors: int = 0

    def to_dict(self) -> dict[str, float | int]:
        """Convert statistics to a dictionary."""
        return {
            "created_at": self.created_at,
            "topics_subscribed": self.topics_subscribed,
            "messages_published": self.messages_published,
            "messages_received": self.messages_received,
            "subscription_errors": self.subscription_errors,
        }


class Subscription:
    """Represents a subscription to a single topic."""

    def __init__(self, *, topic: str, manager: PubSubManager, max_queue_size: int) -> None:
        """Initialize the Subscription."""
        self._topic = topic
        self._manager = manager
        self._max_queue_size = max_queue_size
        self._queue: asyncio.Queue[bytes | None] | None = None

    async def __aenter__(self) -> Self:
        """Enter the async context and initialize the message queue."""
        if self._queue is None:
            self._queue = asyncio.Queue(maxsize=self._max_queue_size)
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and unsubscribe."""
        await self.unsubscribe()

    async def unsubscribe(self) -> None:
        """Unsubscribe from the topic."""
        await self._manager.unsubscribe(topic=self._topic)

    def __aiter__(self) -> AsyncIterator[bytes]:
        """Return self as the asynchronous iterator."""
        return self

    async def __anext__(self) -> bytes:
        """Get the next message in the async iteration."""
        if self._queue is None:
            raise PubSubError(
                message=(
                    "Subscription has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        try:
            message = await self._queue.get()
            if message is None:
                raise StopAsyncIteration
            return message
        except asyncio.CancelledError:
            raise StopAsyncIteration


class PubSubManager:
    """Manages the Pub/Sub lifecycle over a single WebTransport session."""

    def __init__(self, *, stream: WebTransportStream, max_queue_size: int) -> None:
        """Initialize the PubSubManager."""
        self._stream: WebTransportStream = stream
        self._default_max_queue_size = max_queue_size
        self._ingress_task: asyncio.Task[None] | None = None
        self._lock: asyncio.Lock | None = None
        self._subscriptions: dict[str, set[Subscription]] = defaultdict(set)
        self._pending_subscriptions: dict[str, tuple[asyncio.Future[Subscription], Subscription]] = {}
        self._stats = PubSubStats()
        self._is_closing = False

    @property
    def stats(self) -> PubSubStats:
        """Get the current statistics for the Pub/Sub manager."""
        return self._stats

    async def __aenter__(self) -> Self:
        """Enter the async context, ensuring the manager is initialized."""
        if self._lock is None:
            self._lock = asyncio.Lock()
        if self._ingress_task is None:
            self._ingress_task = asyncio.create_task(coro=self._ingress_loop())
            self._ingress_task.add_done_callback(self._on_ingress_done)
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context, closing the manager."""
        await self.close()

    async def close(self) -> None:
        """Close the Pub/Sub manager and its underlying resources."""
        if self._is_closing:
            return

        self._is_closing = True

        if self._ingress_task and not self._ingress_task.done():
            self._ingress_task.cancel()
        if self._stream and not self._stream.is_closed:
            await self._stream.close(error_code=ErrorCodes.APPLICATION_ERROR)

        for future, _ in self._pending_subscriptions.values():
            if not future.done():
                future.set_exception(PubSubError(message="PubSubManager is closing."))
        for subs in self._subscriptions.values():
            for sub in subs:
                if sub._queue and not sub._queue.full():
                    sub._queue.put_nowait(item=None)

        self._pending_subscriptions.clear()
        self._subscriptions.clear()

    async def publish(self, *, topic: str, data: Data) -> None:
        """Publish a message to a topic."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._stream.is_closed:
                raise PubSubError(message="Pub/Sub stream is not available.")

            data_bytes = ensure_bytes(data=data)
            command = b"PUB %b %d\n" % (topic.encode(), len(data_bytes))

            await self._stream.write(data=command + data_bytes)
            self._stats.messages_published += 1

    async def subscribe(self, *, topic: str, timeout: float = 10.0, max_queue_size: int | None = None) -> Subscription:
        """Subscribe to a topic and return a subscription object."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        actual_queue_size = max_queue_size if max_queue_size is not None else self._default_max_queue_size
        if actual_queue_size <= 0:
            raise ValueError("max_queue_size must be positive.")

        async with self._lock:
            if topic in self._pending_subscriptions:
                raise PubSubError(message=f"Subscription to topic '{topic}' is already pending.")

            future: asyncio.Future[Subscription] = asyncio.get_running_loop().create_future()
            subscription = Subscription(topic=topic, manager=self, max_queue_size=actual_queue_size)
            self._pending_subscriptions[topic] = (future, subscription)

        if self._stream.is_closed:
            raise PubSubError(message="Pub/Sub stream is not available.")
        await self._stream.write(data=b"SUB %b\n" % topic.encode())

        try:
            async with asyncio.timeout(delay=timeout):
                confirmed_subscription = await future
            async with self._lock:
                self._subscriptions[topic].add(confirmed_subscription)
                self._stats.topics_subscribed = len(self._subscriptions)
            return confirmed_subscription
        except asyncio.TimeoutError:
            self._stats.subscription_errors += 1
            async with self._lock:
                self._pending_subscriptions.pop(topic, None)
            future.cancel()
            raise SubscriptionFailedError(message=f"Subscription to topic '{topic}' timed out.") from None
        except Exception:
            self._stats.subscription_errors += 1
            async with self._lock:
                self._pending_subscriptions.pop(topic, None)
            raise

    async def unsubscribe(self, *, topic: str) -> None:
        """Unsubscribe from a topic."""
        if self._lock is None:
            raise PubSubError(
                message=(
                    "PubSubManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if not self._subscriptions.get(topic):
                return

            if self._stream.is_closed:
                raise NotSubscribedError(message="Pub/Sub stream is not available to unsubscribe.")

            await self._stream.write(data=b"UNSUB %b\n" % topic.encode())

            for sub in self._subscriptions.get(topic, set()):
                if sub._queue and not sub._queue.full():
                    sub._queue.put_nowait(item=None)

            self._subscriptions.pop(topic, None)
            self._stats.topics_subscribed = len(self._subscriptions)

    async def _ingress_loop(self) -> None:
        """Read and dispatch incoming messages and signals."""
        try:
            while not self._stream.is_closed:
                line = await self._stream.readline()
                if not line:
                    break

                parts = line.strip().split(b" ", 2)
                command = parts[0]

                if command == b"MSG" and len(parts) == 3:
                    topic, length_str = parts[1].decode(), parts[2]
                    length = int(length_str)
                    payload = await self._stream.readexactly(n=length)
                    if self._lock:
                        async with self._lock:
                            subs = self._subscriptions.get(topic)
                        if subs:
                            self._stats.messages_received += 1
                            for sub in subs.copy():
                                if sub._queue:
                                    try:
                                        sub._queue.put_nowait(item=payload)
                                    except asyncio.QueueFull:
                                        logger.warning(
                                            "Subscription queue for topic '%s' is full. Dropping message.", topic
                                        )
                elif command == b"SUB-OK" and len(parts) == 2:
                    topic = parts[1].decode()
                    if self._lock:
                        async with self._lock:
                            pending = self._pending_subscriptions.pop(topic, None)
                        if pending:
                            future, subscription = pending
                            if not future.done():
                                future.set_result(subscription)
                elif command == b"SUB-FAIL" and len(parts) == 3:
                    topic, reason = parts[1].decode(), parts[2].decode()
                    if self._lock:
                        async with self._lock:
                            pending = self._pending_subscriptions.pop(topic, None)
                        if pending:
                            future, _ = pending
                            if not future.done():
                                future.set_exception(SubscriptionFailedError(message=reason))
        except (ConnectionError, StreamError, asyncio.IncompleteReadError) as e:
            logger.info("Pub/Sub stream closed: %s", e)
        except Exception as e:
            if not self._is_closing:
                logger.error("Error in Pub/Sub ingress loop: %s", e, exc_info=True)
                raise

    def _on_ingress_done(self, task: asyncio.Task[None]) -> None:
        """Callback to trigger cleanup when the ingress task finishes."""
        if self._is_closing:
            return

        if not task.cancelled():
            if exc := task.exception():
                logger.error("Pub/Sub ingress task finished unexpectedly with an exception: %s.", exc, exc_info=exc)

        if not self._is_closing:
            asyncio.create_task(coro=self.close())



"""Abstractions for the underlying QUIC connection."""

from .connection import ConnectionDiagnostics, WebTransportConnection
from .load_balancer import ConnectionLoadBalancer

__all__: list[str] = ["ConnectionDiagnostics", "ConnectionLoadBalancer", "WebTransportConnection"]



"""Utility for managing a cluster of server instances."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import Any, Self

from pywebtransport.config import ServerConfig
from pywebtransport.exceptions import ServerError
from pywebtransport.server.server import WebTransportServer
from pywebtransport.types import ConnectionState, SessionState
from pywebtransport.utils import get_logger

__all__: list[str] = ["ServerCluster"]

logger = get_logger(name=__name__)


class ServerCluster:
    """Manages the lifecycle of multiple WebTransport server instances."""

    def __init__(self, *, configs: list[ServerConfig]) -> None:
        """Initialize the server cluster."""
        self._configs = configs
        self._servers: list[WebTransportServer] = []
        self._running = False
        self._lock: asyncio.Lock | None = None

    @property
    def is_running(self) -> bool:
        """Check if the cluster is currently running."""
        return self._running

    async def __aenter__(self) -> Self:
        """Enter the async context and start all servers."""
        self._lock = asyncio.Lock()
        await self.start_all()
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        """Exit the async context and stop all servers."""
        await self.stop_all()

    async def start_all(self) -> None:
        """Start all servers in the cluster concurrently."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._running:
                return
            initial_configs = self._configs

        started_servers: list[WebTransportServer] = []
        tasks: list[asyncio.Task[WebTransportServer]] = []
        try:
            async with asyncio.TaskGroup() as tg:
                for config in initial_configs:
                    tasks.append(tg.create_task(coro=self._create_and_start_server(config=config)))
            started_servers = [task.result() for task in tasks]
        except* Exception as eg:
            logger.error("Failed to start server cluster: %s", eg.exceptions, exc_info=True)
            successful_servers = [task.result() for task in tasks if task.done() and not task.exception()]
            if successful_servers:
                logger.info("Cleaning up %d successfully started servers...", len(successful_servers))
                try:
                    async with asyncio.TaskGroup() as cleanup_tg:
                        for server in successful_servers:
                            cleanup_tg.create_task(coro=server.close())
                except* Exception as cleanup_eg:
                    logger.error("Errors during cluster startup cleanup: %s", cleanup_eg.exceptions, exc_info=True)
            raise eg.exceptions[0]

        async with self._lock:
            self._servers = started_servers
            self._running = True
            logger.info("Started cluster with %d servers", len(self._servers))

    async def stop_all(self) -> None:
        """Stop all servers in the cluster concurrently."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        servers_to_stop: list[WebTransportServer] = []
        async with self._lock:
            if not self._running:
                return
            servers_to_stop = self._servers
            self._servers = []
            self._running = False

        if servers_to_stop:
            try:
                async with asyncio.TaskGroup() as tg:
                    for server in servers_to_stop:
                        tg.create_task(coro=server.close())
            except* Exception as eg:
                logger.error("Errors occurred while stopping server cluster: %s", eg.exceptions, exc_info=True)
                raise eg
            logger.info("Stopped server cluster")

    async def add_server(self, *, config: ServerConfig) -> WebTransportServer | None:
        """Add and start a new server in the running cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        is_running: bool
        async with self._lock:
            is_running = self._running
            if not is_running:
                self._configs.append(config)
                logger.info("Cluster not running. Server config added for next start.")
                return None

        try:
            server = await self._create_and_start_server(config=config)
            async with self._lock:
                if not self._running:
                    await server.close()
                    logger.warning("Cluster was stopped while new server was starting. New server has been shut down.")
                    return None
                self._servers.append(server)
                logger.info("Added server to cluster: %s", server.local_address)
            return server
        except Exception as e:
            logger.error("Failed to add server to cluster: %s", e, exc_info=True)
            return None

    async def remove_server(self, *, host: str, port: int) -> bool:
        """Remove and stop a specific server from the cluster by its address."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        server_to_remove: WebTransportServer | None = None
        async with self._lock:
            for server in self._servers:
                if server.local_address == (host, port):
                    server_to_remove = server
                    break

            if server_to_remove:
                self._servers.remove(server_to_remove)
            else:
                logger.warning("Server at %s:%s not found in cluster.", host, port)
                return False

        await server_to_remove.close()
        logger.info("Removed server from cluster: %s:%s", host, port)
        return True

    async def get_cluster_stats(self) -> dict[str, Any]:
        """Get deeply aggregated statistics for the entire cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        servers_snapshot: list[WebTransportServer]
        async with self._lock:
            if not self._servers:
                return {}
            servers_snapshot = self._servers.copy()

        tasks = []
        try:
            async with asyncio.TaskGroup() as tg:
                for s in servers_snapshot:
                    tasks.append(tg.create_task(coro=s.diagnostics()))
        except* Exception as eg:
            logger.error("Failed to fetch stats from some servers: %s", eg.exceptions, exc_info=True)
            raise eg

        diagnostics_list = [task.result() for task in tasks if task.done() and not task.exception()]

        agg_stats: dict[str, Any] = {
            "server_count": len(servers_snapshot),
            "total_connections_accepted": 0,
            "total_connections_rejected": 0,
            "total_connections_active": 0,
            "total_sessions_active": 0,
        }
        for diag in diagnostics_list:
            agg_stats["total_connections_accepted"] += diag.stats.connections_accepted
            agg_stats["total_connections_rejected"] += diag.stats.connections_rejected
            agg_stats["total_connections_active"] += diag.connection_states.get(ConnectionState.CONNECTED, 0)
            agg_stats["total_sessions_active"] += diag.session_states.get(SessionState.CONNECTED, 0)

        return agg_stats

    async def get_server_count(self) -> int:
        """Get the number of running servers in the cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            return len(self._servers)

    async def get_servers(self) -> list[WebTransportServer]:
        """Get a thread-safe copy of all active servers in the cluster."""
        if self._lock is None:
            raise ServerError(
                message=(
                    "ServerCluster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            return self._servers.copy()

    async def _create_and_start_server(self, *, config: ServerConfig) -> WebTransportServer:
        """Create, activate, and start a single server instance."""
        server = WebTransportServer(config=config)
        await server.__aenter__()

        try:
            await server.listen()
        except Exception:
            await server.close()
            raise
        return server
